"""
    Generates sound fields for various arrays and virtual source types.
"""

import numpy as np
import matplotlib.pyplot as plt
import sfs


dx = 0.1  # secondary source distance
N = 30  # number of secondary sources
f = 1000  # frequency
pw_angle = 20  # traveling direction of plane wave
xs = [-1.5, 0.2, 0]  # position of virtual monopole
tapering = sfs.tapering.tukey  # tapering window
talpha = 0.3  # parameter for tapering window
xnorm = [1, 1, 0]  # normalization point for plots
grid = sfs.util.xyz_grid([-2.5, 2.5], [-1.5, 2.5], 0, spacing=0.02)
acenter = [0.3, 0.7, 0]  # center and normal vector of array
anormal = sfs.util.direction_vector(np.radians(35), np.radians(90))

# angular frequency
omega = 2 * np.pi * f
# normal vector of plane wave
npw = sfs.util.direction_vector(np.radians(pw_angle), np.radians(90))


def compute_and_plot_soundfield(title):
    """Compute and plot synthesized sound field."""
    print('Computing', title)

    twin = tapering(selection, alpha=talpha)
    p = sfs.fd.synthesize(d, twin, array, secondary_source, grid=grid)

    plt.figure(figsize=(15, 15))
    plt.cla()
    sfs.plot2d.amplitude(p, grid, xnorm=xnorm)
    sfs.plot2d.loudspeakers(array.x, array.n, twin)
    sfs.plot2d.virtualsource(xs)
    sfs.plot2d.virtualsource([0, 0], npw, type='plane')
    plt.title(title)
    plt.grid()
    plt.savefig(title + '.png')


# linear array, secondary point sources, virtual monopole
array = sfs.array.linear(N, dx, center=acenter, orientation=anormal)

d, selection, secondary_source = sfs.fd.wfs.point_3d(
    omega, array.x, array.n, xs)
compute_and_plot_soundfield('linear_ps_wfs_3d_point')

d, selection, secondary_source = sfs.fd.wfs.point_25d(
    omega, array.x, array.n, xs, xref=xnorm)
compute_and_plot_soundfield('linear_ps_wfs_25d_point')

d, selection, secondary_source = sfs.fd.wfs.point_2d(
    omega, array.x, array.n, xs)
compute_and_plot_soundfield('linear_ps_wfs_2d_point')

# linear array, secondary line sources, virtual line source
d, selection, secondary_source = sfs.fd.wfs.line_2d(
    omega, array.x, array.n, xs)
compute_and_plot_soundfield('linear_ls_wfs_2d_line')


# linear array, secondary point sources, virtual plane wave
d, selection, secondary_source = sfs.fd.wfs.plane_3d(
    omega, array.x, array.n, npw)
compute_and_plot_soundfield('linear_ps_wfs_3d_plane')

d, selection, secondary_source = sfs.fd.wfs.plane_25d(
    omega, array.x, array.n, npw, xref=xnorm)
compute_and_plot_soundfield('linear_ps_wfs_25d_plane')

d, selection, secondary_source = sfs.fd.wfs.plane_2d(
    omega, array.x, array.n, npw)
compute_and_plot_soundfield('linear_ps_wfs_2d_plane')


# non-uniform linear array, secondary point sources
array = sfs.array.linear_diff(N//3 * [dx] + N//3 * [dx/2] + N//3 * [dx],
                                   center=acenter, orientation=anormal)

d, selection, secondary_source = sfs.fd.wfs.point_25d(
    omega, array.x, array.n, xs, xref=xnorm)
compute_and_plot_soundfield('linear_nested_ps_wfs_25d_point')

d, selection, secondary_source = sfs.fd.wfs.plane_25d(
    omega, array.x, array.n, npw, xref=xnorm)
compute_and_plot_soundfield('linear_nested_ps_wfs_25d_plane')


# random sampled linear array, secondary point sources
array = sfs.array.linear_random(N, dx/2, 1.5*dx, center=acenter,
                                     orientation=anormal)

d, selection, secondary_source = sfs.fd.wfs.point_25d(
    omega, array.x, array.n, xs, xref=xnorm)
compute_and_plot_soundfield('linear_random_ps_wfs_25d_point')

d, selection, secondary_source = sfs.fd.wfs.plane_25d(
    omega, array.x, array.n, npw, xref=xnorm)
compute_and_plot_soundfield('linear_random_ps_wfs_25d_plane')


# rectangular array, secondary point sources
array = sfs.array.rectangular((N, N//2), dx, center=acenter, orientation=anormal)
d, selection, secondary_source = sfs.fd.wfs.point_25d(
    omega, array.x, array.n, xs, xref=xnorm)
compute_and_plot_soundfield('rectangular_ps_wfs_25d_point')

d, selection, secondary_source = sfs.fd.wfs.plane_25d(
    omega, array.x, array.n, npw, xref=xnorm)
compute_and_plot_soundfield('rectangular_ps_wfs_25d_plane')


# circular array, secondary point sources
N = 60
array = sfs.array.circular(N, 1, center=acenter)
d, selection, secondary_source = sfs.fd.wfs.point_25d(
    omega, array.x, array.n, xs, xref=xnorm)
compute_and_plot_soundfield('circular_ps_wfs_25d_point')

d, selection, secondary_source = sfs.fd.wfs.plane_25d(
    omega, array.x, array.n, npw, xref=xnorm)
compute_and_plot_soundfield('circular_ps_wfs_25d_plane')


# circular array, secondary line sources, NFC-HOA
array = sfs.array.circular(N, 1)
xnorm = [0, 0, 0]
talpha = 0  # switches off tapering

d, selection, secondary_source = sfs.fd.nfchoa.plane_2d(
    omega, array.x, 1, npw)
compute_and_plot_soundfield('circular_ls_nfchoa_2d_plane')


# circular array, secondary point sources, NFC-HOA
array = sfs.array.circular(N, 1)
xnorm = [0, 0, 0]
talpha = 0  # switches off tapering

d, selection, secondary_source = sfs.fd.nfchoa.point_25d(
    omega, array.x, 1, xs)
compute_and_plot_soundfield('circular_ps_nfchoa_25d_point')

d, selection, secondary_source = sfs.fd.nfchoa.plane_25d(
    omega, array.x, 1, npw)
compute_and_plot_soundfield('circular_ps_nfchoa_25d_plane')



"""Various utility functions.

.. include:: math-definitions.rst

"""

import collections
import numpy as np
from numpy.core.umath_tests import inner1d
from scipy.special import spherical_jn, spherical_yn
from . import default


def rotation_matrix(n1, n2):
    """Compute rotation matrix for rotation from *n1* to *n2*.

    Parameters
    ----------
    n1, n2 : (3,) array_like
        Two vectors.  They don't have to be normalized.

    Returns
    -------
    (3, 3) `numpy.ndarray`
        Rotation matrix.

    """
    n1 = normalize_vector(n1)
    n2 = normalize_vector(n2)
    I = np.identity(3)
    if np.all(n1 == n2):
        return I  # no rotation
    elif np.all(n1 == -n2):
        return -I  # flip
    # TODO: check for *very close to* parallel vectors

    # Algorithm from http://math.stackexchange.com/a/476311
    v = v0, v1, v2 = np.cross(n1, n2)
    s = np.linalg.norm(v)  # sine
    c = np.inner(n1, n2)  # cosine
    vx = [[0, -v2, v1],
          [v2, 0, -v0],
          [-v1, v0, 0]]  # skew-symmetric cross-product matrix
    return I + vx + np.dot(vx, vx) * (1 - c) / s**2


def wavenumber(omega, c=None):
    """Compute the wavenumber for a given radial frequency."""
    if c is None:
        c = default.c
    return omega / c


def direction_vector(alpha, beta=np.pi/2):
    """Compute normal vector from azimuth, colatitude."""
    return sph2cart(alpha, beta, 1)


def sph2cart(alpha, beta, r):
    r"""Spherical to cartesian coordinate transform.

    .. math::

        x = r \cos \alpha \sin \beta \\
        y = r \sin \alpha \sin \beta \\
        z = r \cos \beta

    with :math:`\alpha \in [0, 2\pi), \beta \in [0, \pi], r \geq 0`

    Parameters
    ----------
    alpha : float or array_like
            Azimuth angle in radiants
    beta : float or array_like
            Colatitude angle in radiants (with 0 denoting North pole)
    r : float or array_like
            Radius

    Returns
    -------
    x : float or `numpy.ndarray`
        x-component of Cartesian coordinates
    y : float or `numpy.ndarray`
        y-component of Cartesian coordinates
    z : float or `numpy.ndarray`
        z-component of Cartesian coordinates

    """
    x = r * np.cos(alpha) * np.sin(beta)
    y = r * np.sin(alpha) * np.sin(beta)
    z = r * np.cos(beta)
    return x, y, z


def cart2sph(x, y, z):
    r"""Cartesian to spherical coordinate transform.

    .. math::

        \alpha = \arctan \left( \frac{y}{x} \right) \\
        \beta = \arccos \left( \frac{z}{r} \right) \\
        r = \sqrt{x^2 + y^2 + z^2}

    with :math:`\alpha \in [-pi, pi], \beta \in [0, \pi], r \geq 0`

    Parameters
    ----------
    x : float or array_like
        x-component of Cartesian coordinates
    y : float or array_like
        y-component of Cartesian coordinates
    z : float or array_like
        z-component of Cartesian coordinates

    Returns
    -------
    alpha : float or `numpy.ndarray`
            Azimuth angle in radiants
    beta : float or `numpy.ndarray`
            Colatitude angle in radiants (with 0 denoting North pole)
    r : float or `numpy.ndarray`
            Radius

    """
    r = np.sqrt(x**2 + y**2 + z**2)
    alpha = np.arctan2(y, x)
    beta = np.arccos(z / r)
    return alpha, beta, r


def asarray_1d(a, **kwargs):
    """Squeeze the input and check if the result is one-dimensional.

    Returns *a* converted to a `numpy.ndarray` and stripped of
    all singleton dimensions.  Scalars are "upgraded" to 1D arrays.
    The result must have exactly one dimension.
    If not, an error is raised.

    """
    result = np.squeeze(np.asarray(a, **kwargs))
    if result.ndim == 0:
        result = result.reshape((1,))
    elif result.ndim > 1:
        raise ValueError("array must be one-dimensional")
    return result


def asarray_of_rows(a, **kwargs):
    """Convert to 2D array, turn column vector into row vector.

    Returns *a* converted to a `numpy.ndarray` and stripped of
    all singleton dimensions.  If the result has exactly one dimension,
    it is re-shaped into a 2D row vector.

    """
    result = np.squeeze(np.asarray(a, **kwargs))
    if result.ndim == 1:
        result = result.reshape(1, -1)
    return result


def as_xyz_components(components, **kwargs):
    r"""Convert *components* to `XyzComponents` of `numpy.ndarray`\s.

    The *components* are first converted to NumPy arrays (using
    :func:`numpy.asarray`) which are then assembled into an
    `XyzComponents` object.

    Parameters
    ----------
    components : triple or pair of array_like
        The values to be used as X, Y and Z arrays.  Z is optional.
    **kwargs
        All further arguments are forwarded to :func:`numpy.asarray`,
        which is applied to the elements of *components*.

    """
    return XyzComponents([np.asarray(c, **kwargs) for c in components])


def as_delayed_signal(arg, **kwargs):
    """Make sure that the given argument can be used as a signal.

    Parameters
    ----------
    arg : sequence of 1 array_like followed by 1 or 2 scalars
        The first element is converted to a NumPy array, the second
        element is used as the sampling rate (in Hertz) and the optional
        third element is used as the starting time of the signal (in
        seconds).  Default starting time is 0.
    **kwargs
        All keyword arguments are forwarded to :func:`numpy.asarray`.

    Returns
    -------
    `DelayedSignal`
        A named tuple consisting of a `numpy.ndarray` containing the
        audio data, followed by the sampling rate (in Hertz) and the
        starting time (in seconds) of the signal.

    Examples
    --------
    Typically, this is used together with tuple unpacking to assign the
    audio data, the sampling rate and the starting time to separate
    variables:

    >>> import sfs
    >>> sig = [1], 44100
    >>> data, fs, signal_offset = sfs.util.as_delayed_signal(sig)
    >>> data
    array([1])
    >>> fs
    44100
    >>> signal_offset
    0

    """
    try:
        data, samplerate, *time = arg
        time, = time or [0]
    except (IndexError, TypeError, ValueError):
        pass
    else:
        valid_arguments = (not np.isscalar(data) and
                           np.isscalar(samplerate) and
                           np.isscalar(time))
        if valid_arguments:
            data = np.asarray(data, **kwargs)
            return DelayedSignal(data, samplerate, time)
    raise TypeError('expected audio data, samplerate, optional start time')


def strict_arange(start, stop, step=1, *, endpoint=False, dtype=None,
                  **kwargs):
    """Like :func:`numpy.arange`, but compensating numeric errors.

    Unlike :func:`numpy.arange`, but similar to :func:`numpy.linspace`,
    providing ``endpoint=True`` includes both endpoints.

    Parameters
    ----------
    start, stop, step, dtype
        See :func:`numpy.arange`.
    endpoint
        See :func:`numpy.linspace`.

        .. note:: With ``endpoint=True``, the difference between *start*
           and *end* value must be an integer multiple of the
           corresponding *spacing* value!
    **kwargs
        All further arguments are forwarded to :func:`numpy.isclose`.

    Returns
    -------
    `numpy.ndarray`
        Array of evenly spaced values.  See :func:`numpy.arange`.

    """
    remainder = (stop - start) % step
    if np.any(np.isclose(remainder, (0.0, step), **kwargs)):
        if endpoint:
            stop += step * 0.5
        else:
            stop -= step * 0.5
    elif endpoint:
        raise ValueError("Invalid stop value for endpoint=True")
    return np.arange(start, stop, step, dtype)


def xyz_grid(x, y, z, *, spacing, endpoint=True, **kwargs):
    """Create a grid with given range and spacing.

    Parameters
    ----------
    x, y, z : float or pair of float
        Inclusive range of the respective coordinate or a single value
        if only a slice along this dimension is needed.
    spacing : float or triple of float
        Grid spacing.  If a single value is specified, it is used for
        all dimensions, if multiple values are given, one value is used
        per dimension.  If a dimension (*x*, *y* or *z*) has only a
        single value, the corresponding spacing is ignored.
    endpoint : bool, optional
        If ``True`` (the default), the endpoint of each range is
        included in the grid.  Use ``False`` to get a result similar to
        :func:`numpy.arange`.  See `strict_arange()`.
    **kwargs
        All further arguments are forwarded to `strict_arange()`.

    Returns
    -------
    `XyzComponents`
        A grid that can be used for sound field calculations.

    See Also
    --------
    strict_arange, numpy.meshgrid

    """
    if np.isscalar(spacing):
        spacing = [spacing] * 3
    ranges = []
    scalars = []
    for i, coord in enumerate([x, y, z]):
        if np.isscalar(coord):
            scalars.append((i, coord))
        else:
            start, stop = coord
            ranges.append(strict_arange(start, stop, spacing[i],
                                        endpoint=endpoint, **kwargs))
    grid = np.meshgrid(*ranges, sparse=True, copy=False)
    for i, s in scalars:
        grid.insert(i, s)
    return XyzComponents(grid)


def normalize(p, grid, xnorm):
    """Normalize sound field wrt position *xnorm*."""
    return p / np.abs(probe(p, grid, xnorm))


def probe(p, grid, x):
    """Determine the value at position *x* in the sound field *p*."""
    grid = as_xyz_components(grid)
    x = asarray_1d(x)
    r = np.linalg.norm(grid - x)
    idx = np.unravel_index(r.argmin(), r.shape)
    return p[idx]


def broadcast_zip(*args):
    """Broadcast arguments to the same shape and then use :func:`zip`."""
    return zip(*np.broadcast_arrays(*args))


def normalize_vector(x):
    """Normalize a 1D vector."""
    x = asarray_1d(x)
    return x / np.linalg.norm(x)


def db(x, *, power=False):
    """Convert *x* to decibel.

    Parameters
    ----------
    x : array_like
        Input data.  Values of 0 lead to negative infinity.
    power : bool, optional
        If ``power=False`` (the default), *x* is squared before
        conversion.

    """
    with np.errstate(divide='ignore'):
        return (10 if power else 20) * np.log10(np.abs(x))


class XyzComponents(np.ndarray):
    """See __init__()."""

    def __init__(self, components):
        r"""Triple (or pair) of components: x, y, and optionally z.

        Instances of this class can be used to store coordinate grids
        (either regular grids like in `xyz_grid()` or arbitrary point
        clouds) or vector fields (e.g. particle velocity).

        This class is a subclass of `numpy.ndarray`.  It is
        one-dimensional (like a plain `list`) and has a length of 3 (or
        2, if no z-component is available).  It uses ``dtype=object`` in
        order to be able to store other `numpy.ndarray`\s of arbitrary
        shapes but also scalars, if needed.  Because it is a NumPy array
        subclass, it can be used in operations with scalars and "normal"
        NumPy arrays, as long as they have a compatible shape.  Like any
        NumPy array, instances of this class are iterable and can be
        used, e.g., in for-loops and tuple unpacking.  If slicing or
        broadcasting leads to an incompatible shape, a plain
        `numpy.ndarray` with ``dtype=object`` is returned.

        To make sure the *components* are NumPy arrays themselves, use
        `as_xyz_components()`.

        Parameters
        ----------
        components : (3,) or (2,) array_like
            The values to be used as X, Y and Z data.  Z is optional.

        """
        # This method does nothing, it's only here for the documentation!

    def __new__(cls, components):
        # object arrays cannot be created and populated in a single step:
        obj = np.ndarray.__new__(cls, len(components), dtype=object)
        for i, component in enumerate(components):
            obj[i] = component
        return obj

    def __array_finalize__(self, obj):
        if self.ndim == 0:
            pass  # this is allowed, e.g. for np.inner()
        elif self.ndim > 1 or len(self) not in (2, 3):
            raise ValueError("XyzComponents can only have 2 or 3 components")

    def __array_prepare__(self, obj, context=None):
        if obj.ndim == 1 and len(obj) in (2, 3):
            return obj.view(XyzComponents)
        return obj

    def __array_wrap__(self, obj, context=None):
        if obj.ndim != 1 or len(obj) not in (2, 3):
            return obj.view(np.ndarray)
        return obj

    def __getitem__(self, index):
        if isinstance(index, slice):
            start, stop, step = index.indices(len(self))
            if start == 0 and stop in (2, 3) and step == 1:
                return np.ndarray.__getitem__(self, index)
        # Slices other than xy and xyz are "downgraded" to ndarray
        return np.ndarray.__getitem__(self.view(np.ndarray), index)

    def __repr__(self):
        return 'XyzComponents(\n' + ',\n'.join(
            '    {}={}'.format(name, repr(data).replace('\n', '\n      '))
            for name, data in zip('xyz', self)) + ')'

    def make_property(index, doc):

        def getter(self):
            return self[index]

        def setter(self, value):
            self[index] = value

        return property(getter, setter, doc=doc)

    x = make_property(0, doc='x-component.')
    y = make_property(1, doc='y-component.')
    z = make_property(2, doc='z-component (optional).')

    del make_property

    def apply(self, func, *args, **kwargs):
        """Apply a function to each component.

        The function *func* will be called once for each component,
        passing the current component as first argument.  All further
        arguments are passed after that.
        The results are returned as a new `XyzComponents` object.

        """
        return XyzComponents([func(i, *args, **kwargs) for i in self])


DelayedSignal = collections.namedtuple('DelayedSignal', 'data samplerate time')
"""A tuple of audio data, sampling rate and start time.

This class (a `collections.namedtuple`) is not meant to be instantiated
by users.

To pass a signal to a function, just use a simple `tuple` or `list`
containing the audio data and the sampling rate (in Hertz), with an
optional starting time (in seconds) as a third item.
If you want to ensure that a given variable contains a valid signal, use
`sfs.util.as_delayed_signal()`.

"""


def image_sources_for_box(x, L, N, *, prune=True):
    """Image source method for a cuboid room.

    The classical method by Allen and Berkley :cite:`Allen1979`.

    Parameters
    ----------
    x : (D,) array_like
        Original source location within box.
        Values between 0 and corresponding side length.
    L : (D,) array_like
        side lengths of room.
    N : int
        Maximum number of reflections per image source, see below.
    prune : bool, optional
        selection of image sources:

        - If True (default):
          Returns all images reflected up to N times.
          This is the usual interpretation of N as "maximum order".

        - If False:
          Returns reflected up to N times between individual wall pairs,
          a total number of :math:`M := (2N+1)^D`.
          This larger set is useful e.g. to select image sources based on
          distance to listener, as suggested by :cite:`Borish1984`.


    Returns
    -------
    xs : (M, D) `numpy.ndarray`
        original & image source locations.
    wall_count : (M, 2D) `numpy.ndarray`
        number of reflections at individual walls for each source.

    """
    def _images_1d_unit_box(x, N):
        result = np.arange(-N, N + 1, dtype=x.dtype)
        result[N % 2::2] += x
        result[1 - (N % 2)::2] += 1 - x
        return result

    def _count_walls_1d(a):
        b = np.floor(a/2)
        c = np.ceil((a-1)/2)
        return np.abs(np.stack([b, c], axis=1)).astype(int)

    L = asarray_1d(L)
    x = asarray_1d(x)/L
    D = len(x)
    xs = [_images_1d_unit_box(coord, N) for coord in x]
    xs = np.reshape(np.transpose(np.meshgrid(*xs, indexing='ij')), (-1, D))

    wall_count = np.concatenate([_count_walls_1d(d) for d in xs.T], axis=1)
    xs *= L

    if prune is True:
        N_mask = np.sum(wall_count, axis=1) <= N
        xs = xs[N_mask, :]
        wall_count = wall_count[N_mask, :]

    return xs, wall_count


def spherical_hn2(n, z):
    r"""Spherical Hankel function of 2nd kind.

    Defined as https://dlmf.nist.gov/10.47.E6,

    .. math::

        \hankel{2}{n}{z} = \sqrt{\frac{\pi}{2z}}
        \Hankel{2}{n + \frac{1}{2}}{z},

    where :math:`\Hankel{2}{n}{\cdot}` is the Hankel function of the
    second kind and n-th order, and :math:`z` its complex argument.

    Parameters
    ----------
    n : array_like
        Order of the spherical Hankel function (n >= 0).
    z : array_like
        Argument of the spherical Hankel function.

    """
    return spherical_jn(n, z) - 1j * spherical_yn(n, z)


def source_selection_plane(n0, n):
    """Secondary source selection for a plane wave.

    Eq.(13) from :cite:`Spors2008`

    """
    n0 = asarray_of_rows(n0)
    n = normalize_vector(n)
    return np.inner(n, n0) >= default.selection_tolerance


def source_selection_point(n0, x0, xs):
    """Secondary source selection for a point source.

    Eq.(15) from :cite:`Spors2008`

    """
    n0 = asarray_of_rows(n0)
    x0 = asarray_of_rows(x0)
    xs = asarray_1d(xs)
    ds = x0 - xs
    return inner1d(ds, n0) >= default.selection_tolerance


def source_selection_line(n0, x0, xs):
    """Secondary source selection for a line source.

    compare Eq.(15) from :cite:`Spors2008`

    """
    return source_selection_point(n0, x0, xs)


def source_selection_focused(ns, x0, xs):
    """Secondary source selection for a focused source.

    Eq.(2.78) from :cite:`Wierstorf2014`

    """
    x0 = asarray_of_rows(x0)
    xs = asarray_1d(xs)
    ns = normalize_vector(ns)
    ds = xs - x0
    return inner1d(ns, ds) >= default.selection_tolerance


def source_selection_all(N):
    """Select all secondary sources."""
    return np.ones(N, dtype=bool)


def max_order_circular_harmonics(N):
    r"""Maximum order of 2D/2.5D HOA.

    It returns the maximum order for which no spatial aliasing appears.
    It is given on page 132 of [Ahrens2012]_ as

    .. math::
        \mathtt{max\_order} =
            \begin{cases}
                N/2 - 1 & \text{even}\;N \\
                (N-1)/2 & \text{odd}\;N,
            \end{cases}

    which is equivalent to

    .. math::
        \mathtt{max\_order} = \big\lfloor \frac{N - 1}{2} \big\rfloor.

    Parameters
    ----------
    N : int
        Number of secondary sources.

    """
    return (N - 1) // 2


def max_order_spherical_harmonics(N):
    r"""Maximum order of 3D HOA.

    .. math::
        \mathtt{max\_order} = \lfloor \sqrt{N} \rfloor - 1.

    Parameters
    ----------
    N : int
        Number of secondary sources.

    """
    return int(np.sqrt(N) - 1)



"""Compute ESA driving functions for various systems.

ESA is abbreviation for equivalent scattering approach.

ESA driving functions for an edge-shaped SSD are provided below.
Further ESA for different geometries might be added here.

Note that mode-matching (such as NFC-HOA, SDM) are equivalent
to ESA in their specific geometries (spherical/circular, planar/linear).

"""

import numpy as np
from scipy.special import jn, hankel2
from .. import util
from . import secondary_source_line, secondary_source_point


def plane_2d_edge(omega, x0, n=[0, 1, 0], *, alpha=3/2*np.pi, Nc=None, c=None):
    r"""Driving function for 2-dimensional plane wave with edge ESA.

    Driving function for a virtual plane wave using the 2-dimensional ESA
    for an edge-shaped secondary source distribution consisting of
    monopole line sources.

    Parameters
    ----------
    omega : float
        Angular frequency.
    x0 : int(N, 3) array_like
        Sequence of secondary source positions.
    n : (3,) array_like, optional
        Normal vector of synthesized plane wave.
    alpha : float, optional
        Outer angle of edge.
    Nc : int, optional
        Number of elements for series expansion of driving function. Estimated
        if not given.
    c : float, optional
        Speed of sound

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    One leg of the secondary sources has to be located on the x-axis (y0=0),
    the edge at the origin.

    Derived from :cite:`Spors2016`

    """
    x0 = np.asarray(x0)
    n = util.normalize_vector(n)
    k = util.wavenumber(omega, c)
    phi_s = np.arctan2(n[1], n[0]) + np.pi
    L = x0.shape[0]

    r = np.linalg.norm(x0, axis=1)
    phi = np.arctan2(x0[:, 1], x0[:, 0])
    phi = np.where(phi < 0, phi+2*np.pi, phi)

    if Nc is None:
        Nc = np.ceil(2 * k * np.max(r) * alpha/np.pi)

    epsilon = np.ones(Nc)  # weights for series expansion
    epsilon[0] = 2

    d = np.zeros(L, dtype=complex)
    for m in np.arange(Nc):
        nu = m*np.pi/alpha
        d = d + 1/epsilon[m] * np.exp(1j*nu*np.pi/2) * np.sin(nu*phi_s) \
            * np.cos(nu*phi) * nu/r * jn(nu, k*r)

    d[phi > 0] = -d[phi > 0]

    selection = util.source_selection_all(len(x0))
    return 4*np.pi/alpha * d, selection, secondary_source_line(omega, c)


def plane_2d_edge_dipole_ssd(omega, x0, n=[0, 1, 0], *, alpha=3/2*np.pi,
                             Nc=None, c=None):
    r"""Driving function for 2-dimensional plane wave with edge dipole ESA.

    Driving function for a virtual plane wave using the 2-dimensional ESA
    for an edge-shaped secondary source distribution consisting of
    dipole line sources.

    Parameters
    ----------
    omega : float
        Angular frequency.
    x0 : int(N, 3) array_like
        Sequence of secondary source positions.
    n : (3,) array_like, optional
        Normal vector of synthesized plane wave.
    alpha : float, optional
        Outer angle of edge.
    Nc : int, optional
        Number of elements for series expansion of driving function. Estimated
        if not given.
    c : float, optional
        Speed of sound

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    One leg of the secondary sources has to be located on the x-axis (y0=0),
    the edge at the origin.

    Derived from :cite:`Spors2016`

    """
    x0 = np.asarray(x0)
    n = util.normalize_vector(n)
    k = util.wavenumber(omega, c)
    phi_s = np.arctan2(n[1], n[0]) + np.pi
    L = x0.shape[0]

    r = np.linalg.norm(x0, axis=1)
    phi = np.arctan2(x0[:, 1], x0[:, 0])
    phi = np.where(phi < 0, phi+2*np.pi, phi)

    if Nc is None:
        Nc = np.ceil(2 * k * np.max(r) * alpha/np.pi)

    epsilon = np.ones(Nc)  # weights for series expansion
    epsilon[0] = 2

    d = np.zeros(L, dtype=complex)
    for m in np.arange(Nc):
        nu = m*np.pi/alpha
        d = d + 1/epsilon[m] * np.exp(1j*nu*np.pi/2) * np.cos(nu*phi_s) \
            * np.cos(nu*phi) * jn(nu, k*r)

    return 4*np.pi/alpha * d


def line_2d_edge(omega, x0, xs, *, alpha=3/2*np.pi, Nc=None, c=None):
    r"""Driving function for 2-dimensional line source with edge ESA.

    Driving function for a virtual line source using the 2-dimensional ESA
    for an edge-shaped secondary source distribution consisting of line
    sources.

    Parameters
    ----------
    omega : float
        Angular frequency.
    x0 : int(N, 3) array_like
        Sequence of secondary source positions.
    xs : (3,) array_like
        Position of synthesized line source.
    alpha : float, optional
        Outer angle of edge.
    Nc : int, optional
        Number of elements for series expansion of driving function. Estimated
        if not given.
    c : float, optional
        Speed of sound

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    One leg of the secondary sources has to be located on the x-axis (y0=0),
    the edge at the origin.

    Derived from :cite:`Spors2016`

    """
    x0 = np.asarray(x0)
    k = util.wavenumber(omega, c)
    phi_s = np.arctan2(xs[1], xs[0])
    if phi_s < 0:
        phi_s = phi_s + 2*np.pi
    r_s = np.linalg.norm(xs)
    L = x0.shape[0]

    r = np.linalg.norm(x0, axis=1)
    phi = np.arctan2(x0[:, 1], x0[:, 0])
    phi = np.where(phi < 0, phi+2*np.pi, phi)

    if Nc is None:
        Nc = np.ceil(2 * k * np.max(r) * alpha/np.pi)

    epsilon = np.ones(Nc)  # weights for series expansion
    epsilon[0] = 2

    d = np.zeros(L, dtype=complex)
    idx = (r <= r_s)
    for m in np.arange(Nc):
        nu = m*np.pi/alpha
        f = 1/epsilon[m] * np.sin(nu*phi_s) * np.cos(nu*phi) * nu/r
        d[idx] = d[idx] + f[idx] * jn(nu, k*r[idx]) * hankel2(nu, k*r_s)
        d[~idx] = d[~idx] + f[~idx] * jn(nu, k*r_s) * hankel2(nu, k*r[~idx])

    d[phi > 0] = -d[phi > 0]

    selection = util.source_selection_all(len(x0))
    return -1j*np.pi/alpha * d, selection, secondary_source_line(omega, c)


def line_2d_edge_dipole_ssd(omega, x0, xs, *, alpha=3/2*np.pi, Nc=None,
                            c=None):
    r"""Driving function for 2-dimensional line source with edge dipole ESA.

    Driving function for a virtual line source using the 2-dimensional ESA
    for an edge-shaped secondary source distribution consisting of dipole line
    sources.

    Parameters
    ----------
    omega : float
        Angular frequency.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    xs : (3,) array_like
        Position of synthesized line source.
    alpha : float, optional
        Outer angle of edge.
    Nc : int, optional
        Number of elements for series expansion of driving function. Estimated
        if not given.
    c : float, optional
        Speed of sound

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    One leg of the secondary sources has to be located on the x-axis (y0=0),
    the edge at the origin.

    Derived from :cite:`Spors2016`

    """
    x0 = np.asarray(x0)
    k = util.wavenumber(omega, c)
    phi_s = np.arctan2(xs[1], xs[0])
    if phi_s < 0:
        phi_s = phi_s + 2*np.pi
    r_s = np.linalg.norm(xs)
    L = x0.shape[0]

    r = np.linalg.norm(x0, axis=1)
    phi = np.arctan2(x0[:, 1], x0[:, 0])
    phi = np.where(phi < 0, phi+2*np.pi, phi)

    if Nc is None:
        Nc = np.ceil(2 * k * np.max(r) * alpha/np.pi)

    epsilon = np.ones(Nc)  # weights for series expansion
    epsilon[0] = 2

    d = np.zeros(L, dtype=complex)
    idx = (r <= r_s)
    for m in np.arange(Nc):
        nu = m*np.pi/alpha
        f = 1/epsilon[m] * np.cos(nu*phi_s) * np.cos(nu*phi)
        d[idx] = d[idx] + f[idx] * jn(nu, k*r[idx]) * hankel2(nu, k*r_s)
        d[~idx] = d[~idx] + f[~idx] * jn(nu, k*r_s) * hankel2(nu, k*r[~idx])

    return -1j*np.pi/alpha * d


def point_25d_edge(omega, x0, xs, *, xref=[2, -2, 0], alpha=3/2*np.pi,
                   Nc=None, c=None):
    r"""Driving function for 2.5-dimensional point source with edge ESA.

    Driving function for a virtual point source using the 2.5-dimensional
    ESA for an edge-shaped secondary source distribution consisting of point
    sources.

    Parameters
    ----------
    omega : float
        Angular frequency.
    x0 : int(N, 3) array_like
        Sequence of secondary source positions.
    xs : (3,) array_like
        Position of synthesized line source.
    xref: (3,) array_like or float
        Reference position or reference distance
    alpha : float, optional
        Outer angle of edge.
    Nc : int, optional
        Number of elements for series expansion of driving function. Estimated
        if not given.
    c : float, optional
        Speed of sound

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    One leg of the secondary sources has to be located on the x-axis (y0=0),
    the edge at the origin.

    Derived from :cite:`Spors2016`

    """
    x0 = np.asarray(x0)
    xs = np.asarray(xs)
    xref = np.asarray(xref)

    if np.isscalar(xref):
        a = np.linalg.norm(xref)/np.linalg.norm(xref-xs)
    else:
        a = np.linalg.norm(xref-x0, axis=1)/np.linalg.norm(xref-xs)

    d, selection, _ = line_2d_edge(omega, x0, xs, alpha=alpha, Nc=Nc, c=c)
    return 1j*np.sqrt(a) * d, selection, secondary_source_point(omega, c)



"""Submodules for monochromatic sound fields.

.. autosummary::
    :toctree:

    source

    wfs
    nfchoa
    sdm
    esa

"""
from . import source
from .. import array as _array
from .. import util as _util
import numpy as _np


def shiftphase(p, phase):
    """Shift phase of a sound field."""
    p = _np.asarray(p)
    return p * _np.exp(1j * phase)


def displacement(v, omega):
    r"""Particle displacement.

    .. math::

        d(x, t) = \int_{-\infty}^t v(x, \tau) d\tau

    """
    return _util.as_xyz_components(v) / (1j * omega)


def synthesize(d, weights, ssd, secondary_source_function, **kwargs):
    """Compute sound field for a generic driving function.

    Parameters
    ----------
    d : array_like
        Driving function.
    weights : array_like
        Additional weights applied during integration, e.g. source
        selection and tapering.
    ssd : sequence of between 1 and 3 array_like objects
        Positions, normal vectors and weights of secondary sources.
        A `SecondarySourceDistribution` can also be used.
    secondary_source_function : callable
        A function that generates the sound field of a secondary source.
        This signature is expected::

            secondary_source_function(
                position, normal_vector, weight, driving_function_weight,
                **kwargs) -> numpy.ndarray

    **kwargs
        All keyword arguments are forwarded to *secondary_source_function*.
        This is typically used to pass the *grid* argument.

    """
    ssd = _array.as_secondary_source_distribution(ssd)
    if not (len(ssd.x) == len(ssd.n) == len(ssd.a) == len(d) ==
            len(weights)):
        raise ValueError("length mismatch")
    p = 0
    for x, n, a, d, weight in zip(ssd.x, ssd.n, ssd.a, d, weights):
        if weight != 0:
            p += a * weight * d * secondary_source_function(x, n, **kwargs)
    return p


def secondary_source_point(omega, c):
    """Create a point source for use in `sfs.fd.synthesize()`."""

    def secondary_source(position, _, grid):
        return source.point(omega, position, grid, c=c)

    return secondary_source


def secondary_source_line(omega, c):
    """Create a line source for use in `sfs.fd.synthesize()`."""

    def secondary_source(position, _, grid):
        return source.line(omega, position, grid, c=c)

    return secondary_source


from . import esa
from . import nfchoa
from . import sdm
from . import wfs



"""Compute WFS driving functions.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import matplotlib.pyplot as plt
    import numpy as np
    import sfs
    from scipy.signal import unit_impulse

    # Plane wave
    npw = sfs.util.direction_vector(np.radians(-45))

    # Point source
    xs = -1.5, 1.5, 0
    rs = np.linalg.norm(xs)  # distance from origin
    ts = rs / sfs.default.c  # time-of-arrival at origin

    # Focused source
    xf = -0.5, 0.5, 0
    nf = sfs.util.direction_vector(np.radians(-45))  # normal vector
    rf = np.linalg.norm(xf)  # distance from origin
    tf = rf / sfs.default.c  # time-of-arrival at origin

    # Impulsive excitation
    fs = 44100
    signal = unit_impulse(512), fs

    # Circular loudspeaker array
    N = 32  # number of loudspeakers
    R = 1.5  # radius
    array = sfs.array.circular(N, R)

    grid = sfs.util.xyz_grid([-2, 2], [-2, 2], 0, spacing=0.02)

    def plot(d, selection, secondary_source, t=0):
        p = sfs.td.synthesize(d, selection, array, secondary_source, grid=grid,
                              observation_time=t)
        sfs.plot2d.level(p, grid)
        sfs.plot2d.loudspeakers(array.x, array.n, selection * array.a, size=0.15)

"""
import numpy as np
from numpy.core.umath_tests import inner1d  # element-wise inner product
from .. import default
from .. import util
from . import secondary_source_point, apply_delays


def plane_25d(x0, n0, n=[0, 1, 0], xref=[0, 0, 0], c=None):
    r"""Plane wave model by 2.5-dimensional WFS.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of secondary source orientations.
    n : (3,) array_like, optional
        Normal vector (propagation direction) of synthesized plane wave.
    xref : (3,) array_like, optional
        Reference position
    c : float, optional
        Speed of sound

    Returns
    -------
    delays : (N,) numpy.ndarray
        Delays of secondary sources in seconds.
    weights : (N,) numpy.ndarray
        Weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    Notes
    -----
    2.5D correction factor

    .. math::

        g_0 = \sqrt{2 \pi |x_\mathrm{ref} - x_0|}

    d using a plane wave as source model

    .. math::

        d_{2.5D}(x_0,t) = h(t)
        2 g_0 \scalarprod{n}{n_0}
        \dirac{t - \frac{1}{c} \scalarprod{n}{x_0}}

    with wfs(2.5D) prefilter h(t), which is not implemented yet.

    See :sfs:`d_wfs/#equation-td-wfs-plane-25d`

    Examples
    --------
    .. plot::
        :context: close-figs

        delays, weights, selection, secondary_source = \
            sfs.td.wfs.plane_25d(array.x, array.n, npw)
        d = sfs.td.wfs.driving_signals(delays, weights, signal)
        plot(d, selection, secondary_source)

    """
    if c is None:
        c = default.c
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    n = util.normalize_vector(n)
    xref = util.asarray_1d(xref)
    g0 = np.sqrt(2 * np.pi * np.linalg.norm(xref - x0, axis=1))
    delays = inner1d(n, x0) / c
    weights = 2 * g0 * inner1d(n, n0)
    selection = util.source_selection_plane(n0, n)
    return delays, weights, selection, secondary_source_point(c)


def point_25d(x0, n0, xs, xref=[0, 0, 0], c=None):
    r"""Point source by 2.5-dimensional WFS.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of secondary source orientations.
    xs : (3,) array_like
        Virtual source position.
    xref : (3,) array_like, optional
        Reference position
    c : float, optional
        Speed of sound

    Returns
    -------
    delays : (N,) numpy.ndarray
        Delays of secondary sources in seconds.
    weights: (N,) numpy.ndarray
        Weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    Notes
    -----
    2.5D correction factor

    .. math::

         g_0 = \sqrt{2 \pi |x_\mathrm{ref} - x_0|}


    d using a point source as source model

    .. math::

         d_{2.5D}(x_0,t) = h(t)
         \frac{g_0  \scalarprod{(x_0 - x_s)}{n_0}}
         {2\pi |x_0 - x_s|^{3/2}}
         \dirac{t - \frac{|x_0 - x_s|}{c}}

    with wfs(2.5D) prefilter h(t), which is not implemented yet.

    See :sfs:`d_wfs/#equation-td-wfs-point-25d`

    Examples
    --------
    .. plot::
        :context: close-figs

        delays, weights, selection, secondary_source = \
            sfs.td.wfs.point_25d(array.x, array.n, xs)
        d = sfs.td.wfs.driving_signals(delays, weights, signal)
        plot(d, selection, secondary_source, t=ts)

    """
    if c is None:
        c = default.c
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    xref = util.asarray_1d(xref)
    g0 = np.sqrt(2 * np.pi * np.linalg.norm(xref - x0, axis=1))
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    delays = r/c
    weights = g0 * inner1d(ds, n0) / (2 * np.pi * r**(3/2))
    selection = util.source_selection_point(n0, x0, xs)
    return delays, weights, selection, secondary_source_point(c)


def focused_25d(x0, n0, xs, ns, xref=[0, 0, 0], c=None):
    r"""Point source by 2.5-dimensional WFS.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of secondary source orientations.
    xs : (3,) array_like
        Virtual source position.
    ns : (3,) array_like
        Normal vector (propagation direction) of focused source.
        This is used for secondary source selection,
        see `sfs.util.source_selection_focused()`.
    xref : (3,) array_like, optional
        Reference position
    c : float, optional
        Speed of sound

    Returns
    -------
    delays : (N,) numpy.ndarray
        Delays of secondary sources in seconds.
    weights: (N,) numpy.ndarray
        Weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    Notes
    -----
    2.5D correction factor

    .. math::

         g_0 = \sqrt{\frac{|x_\mathrm{ref} - x_0|}
         {|x_0-x_s| + |x_\mathrm{ref}-x_0|}}


    d using a point source as source model

    .. math::

         d_{2.5D}(x_0,t) = h(t)
         \frac{g_0  \scalarprod{(x_0 - x_s)}{n_0}}
         {|x_0 - x_s|^{3/2}}
         \dirac{t + \frac{|x_0 - x_s|}{c}}

    with wfs(2.5D) prefilter h(t), which is not implemented yet.

    See :sfs:`d_wfs/#equation-td-wfs-focused-25d`

    Examples
    --------
    .. plot::
        :context: close-figs

        delays, weights, selection, secondary_source = \
            sfs.td.wfs.focused_25d(array.x, array.n, xf, nf)
        d = sfs.td.wfs.driving_signals(delays, weights, signal)
        plot(d, selection, secondary_source, t=tf)

    """
    if c is None:
        c = default.c
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    xref = util.asarray_1d(xref)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    g0 = np.sqrt(np.linalg.norm(xref - x0, axis=1)
                 / (np.linalg.norm(xref - x0, axis=1) + r))
    delays = -r/c
    weights = g0 * inner1d(ds, n0) / (2 * np.pi * r**(3/2))
    selection = util.source_selection_focused(ns, x0, xs)
    return delays, weights, selection, secondary_source_point(c)


def driving_signals(delays, weights, signal):
    """Get driving signals per secondary source.

    Returned signals are the delayed and weighted mono input signal
    (with N samples) per channel (C).

    Parameters
    ----------
    delays : (C,) array_like
        Delay in seconds for each channel, negative values allowed.
    weights : (C,) array_like
        Amplitude weighting factor for each channel.
    signal : (N,) array_like + float
        Excitation signal consisting of (mono) audio data and a sampling
        rate (in Hertz).  A `DelayedSignal` object can also be used.

    Returns
    -------
    `DelayedSignal`
        A tuple containing the driving signals (in a `numpy.ndarray`
        with shape ``(N, C)``), followed by the sampling rate (in Hertz)
        and a (possibly negative) time offset (in seconds).

    """
    delays = util.asarray_1d(delays)
    weights = util.asarray_1d(weights)
    data, samplerate, signal_offset = apply_delays(signal, delays)
    return util.DelayedSignal(data * weights, samplerate, signal_offset)



"""Compute NFC-HOA driving functions.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import matplotlib.pyplot as plt
    import numpy as np
    import sfs

    plt.rcParams['figure.figsize'] = 6, 6

    xs = -1.5, 1.5, 0
    # normal vector for plane wave:
    npw = sfs.util.direction_vector(np.radians(-45))
    f = 300  # Hz
    omega = 2 * np.pi * f
    R = 1.5  # Radius of circular loudspeaker array

    grid = sfs.util.xyz_grid([-2, 2], [-2, 2], 0, spacing=0.02)

    array = sfs.array.circular(N=32, R=R)

    def plot(d, selection, secondary_source):
        p = sfs.fd.synthesize(d, selection, array, secondary_source, grid=grid)
        sfs.plot2d.amplitude(p, grid)
        sfs.plot2d.loudspeakers(array.x, array.n, selection * array.a, size=0.15)

"""

import numpy as np
from scipy.special import hankel2
from .. import util
from . import secondary_source_point


def plane_2d(omega, x0, r0, n=[0, 1, 0], *, max_order=None, c=None):
    r"""Driving function for 2-dimensional NFC-HOA for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of plane wave.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of circular secondary source distribution.
    n : (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    max_order : float, optional
        Maximum order of circular harmonics used for the calculation.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\phi_0, \omega) =
        -\frac{2\i}{\pi r_0}
        \sum_{m=-M}^M
        \frac{\i^{-m}}{\Hankel{2}{m}{\wc r_0}}
        \e{\i m (\phi_0 - \phi_\text{pw})}

    See :sfs:`d_nfchoa/#equation-fd-nfchoa-plane-2d`

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.nfchoa.plane_2d(
            omega, array.x, R, npw)
        plot(d, selection, secondary_source)

    """
    if max_order is None:
        max_order = util.max_order_circular_harmonics(len(x0))

    x0 = util.asarray_of_rows(x0)
    k = util.wavenumber(omega, c)
    n = util.normalize_vector(n)
    phi, _, r = util.cart2sph(*n)
    phi0 = util.cart2sph(*x0.T)[0]
    d = 0
    for m in range(-max_order, max_order + 1):
        d += 1j**-m / hankel2(m, k * r0) * np.exp(1j * m * (phi0 - phi))
    selection = util.source_selection_all(len(x0))
    return -2j / (np.pi*r0) * d, selection, secondary_source_point(omega, c)


def point_25d(omega, x0, r0, xs, *, max_order=None, c=None):
    r"""Driving function for 2.5-dimensional NFC-HOA for a virtual point source.

    Parameters
    ----------
    omega : float
        Angular frequency of point source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of circular secondary source distribution.
    xs : (3,) array_like
        Position of point source.
    max_order : float, optional
        Maximum order of circular harmonics used for the calculation.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\phi_0, \omega) =
        \frac{1}{2 \pi r_0}
        \sum_{m=-M}^M
        \frac{\hankel{2}{|m|}{\wc r}}{\hankel{2}{|m|}{\wc r_0}}
        \e{\i m (\phi_0 - \phi)}

    See :sfs:`d_nfchoa/#equation-fd-nfchoa-point-25d`

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.nfchoa.point_25d(
            omega, array.x, R, xs)
        plot(d, selection, secondary_source)

    """
    if max_order is None:
        max_order = util.max_order_circular_harmonics(len(x0))

    x0 = util.asarray_of_rows(x0)
    k = util.wavenumber(omega, c)
    xs = util.asarray_1d(xs)
    phi, _, r = util.cart2sph(*xs)
    phi0 = util.cart2sph(*x0.T)[0]
    hr = util.spherical_hn2(range(0, max_order + 1), k * r)
    hr0 = util.spherical_hn2(range(0, max_order + 1), k * r0)
    d = 0
    for m in range(-max_order, max_order + 1):
        d += hr[abs(m)] / hr0[abs(m)] * np.exp(1j * m * (phi0 - phi))
    selection = util.source_selection_all(len(x0))
    return d / (2 * np.pi * r0), selection, secondary_source_point(omega, c)


def plane_25d(omega, x0, r0, n=[0, 1, 0], *, max_order=None, c=None):
    r"""Driving function for 2.5-dimensional NFC-HOA for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of point source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of circular secondary source distribution.
    n : (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    max_order : float, optional
        Maximum order of circular harmonics used for the calculation.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\phi_0, \omega) =
        \frac{2\i}{r_0}
        \sum_{m=-M}^M
        \frac{\i^{-|m|}}{\wc \hankel{2}{|m|}{\wc r_0}}
        \e{\i m (\phi_0 - \phi_\text{pw})}

    See :sfs:`d_nfchoa/#equation-fd-nfchoa-plane-25d`

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.nfchoa.plane_25d(
            omega, array.x, R, npw)
        plot(d, selection, secondary_source)

    """
    if max_order is None:
        max_order = util.max_order_circular_harmonics(len(x0))

    x0 = util.asarray_of_rows(x0)
    k = util.wavenumber(omega, c)
    n = util.normalize_vector(n)
    phi, _, r = util.cart2sph(*n)
    phi0 = util.cart2sph(*x0.T)[0]
    d = 0
    hn2 = util.spherical_hn2(range(0, max_order + 1), k * r0)
    for m in range(-max_order, max_order + 1):
        d += (-1j)**abs(m) / (k * hn2[abs(m)]) * np.exp(1j * m * (phi0 - phi))
    selection = util.source_selection_all(len(x0))
    return 2*1j / r0 * d, selection, secondary_source_point(omega, c)



"""Compute WFS driving functions.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import matplotlib.pyplot as plt
    import numpy as np
    import sfs

    plt.rcParams['figure.figsize'] = 6, 6

    xs = -1.5, 1.5, 0
    xs_focused = -0.5, 0.5, 0
    # normal vector for plane wave:
    npw = sfs.util.direction_vector(np.radians(-45))
    # normal vector for focused source:
    ns_focused = sfs.util.direction_vector(np.radians(-45))
    f = 300  # Hz
    omega = 2 * np.pi * f
    R = 1.5  # Radius of circular loudspeaker array

    grid = sfs.util.xyz_grid([-2, 2], [-2, 2], 0, spacing=0.02)

    array = sfs.array.circular(N=32, R=R)

    def plot(d, selection, secondary_source):
        p = sfs.fd.synthesize(d, selection, array, secondary_source, grid=grid)
        sfs.plot2d.amplitude(p, grid)
        sfs.plot2d.loudspeakers(array.x, array.n, selection * array.a, size=0.15)

"""

import numpy as np
from numpy.core.umath_tests import inner1d  # element-wise inner product
from scipy.special import hankel2
from .. import util
from . import secondary_source_line, secondary_source_point


def line_2d(omega, x0, n0, xs, *, c=None):
    r"""Driving function for 2-dimensional WFS for a virtual line source.

    Parameters
    ----------
    omega : float
        Angular frequency of line source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of virtual line source.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\x_0,\w) = \frac{\i}{2} \wc
            \frac{\scalarprod{\x-\x_0}{\n_0}}{|\x-\x_\text{s}|}
            \Hankel{2}{1}{\wc|\x-\x_\text{s}|}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.line_2d(
            omega, array.x, array.n, xs)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = -1j/2 * k * inner1d(ds, n0) / r * hankel2(1, k * r)
    selection = util.source_selection_line(n0, x0, xs)
    return d, selection, secondary_source_line(omega, c)


def _point(omega, x0, n0, xs, *, c=None):
    r"""Driving function for 2/3-dimensional WFS for a virtual point source.

    Parameters
    ----------
    omega : float
        Angular frequency of point source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of virtual point source.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\x_0, \w) = \i\wc \frac{\scalarprod{\x_0-\x_\text{s}}{\n_0}}
            {|\x_0-\x_\text{s}|^{\frac{3}{2}}}
            \e{-\i\wc |\x_0-\x_\text{s}|}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.point_3d(
            omega, array.x, array.n, xs)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = 1j * k * inner1d(ds, n0) / r ** (3 / 2) * np.exp(-1j * k * r)
    selection = util.source_selection_point(n0, x0, xs)
    return d, selection, secondary_source_point(omega, c)


point_2d = _point


def point_25d(omega, x0, n0, xs, xref=[0, 0, 0], c=None, omalias=None):
    r"""Driving function for 2.5-dimensional WFS of a virtual point source.

    .. versionchanged:: 0.5
        see notes, old handling of `point_25d()` is now `point_25d_legacy()`

    Parameters
    ----------
    omega : float
        Angular frequency of point source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of virtual point source.
    xref : (3,) array_like, optional
        Reference point xref or contour xref(x0) for amplitude correct
        synthesis.
    c : float, optional
        Speed of sound in m/s.
    omalias: float, optional
        Angular frequency where spatial aliasing becomes prominent.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    `point_25d()` derives 2.5D WFS from the 3D
    Neumann-Rayleigh integral (i.e. the TU Delft approach).
    The eq. (3.10), (3.11) in :cite:`Start1997`, equivalent to
    Eq. (2.137) in :cite:`Schultz2016`

    .. math::

        D(\x_0,\w) = \sqrt{8 \pi \, \i\wc}
            \sqrt{\frac{|\x_\text{ref}-\x_0| \cdot
            |\x_0-\x_\text{s}|}{|\x_\text{ref}-\x_0| + |\x_0-\x_\text{s}|}}
            \scalarprod{\frac{\x_0-\x_\text{s}}{|\x_0-\x_\text{s}|}}{\n_0}
            \frac{\e{-\i\wc |\x_0-\x_\text{s}|}}{4\pi\,|\x_0-\x_\text{s}|}

    is implemented.
    The theoretical link of `point_25d()` and `point_25d_legacy()` was
    introduced as *unified WFS framework* in :cite:`Firtha2017`.

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.point_25d(
            omega, array.x, array.n, xs)
        normalize_gain = 4 * np.pi * np.linalg.norm(xs)
        plot(normalize_gain * d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    xref = util.asarray_1d(xref)
    k = util.wavenumber(omega, c)

    ds = x0 - xs
    dr = xref - x0
    s = np.linalg.norm(ds, axis=1)
    r = np.linalg.norm(dr, axis=1)

    d = (
        preeq_25d(omega, omalias, c) *
        np.sqrt(8 * np.pi) *
        np.sqrt((r * s) / (r + s)) *
        inner1d(n0, ds) / s *
        np.exp(-1j * k * s) / (4 * np.pi * s))
    selection = util.source_selection_point(n0, x0, xs)
    return d, selection, secondary_source_point(omega, c)


point_3d = _point


def point_25d_legacy(omega, x0, n0, xs, xref=[0, 0, 0], c=None, omalias=None):
    r"""Driving function for 2.5-dimensional WFS for a virtual point source.

    .. versionadded:: 0.5
        `point_25d()` was renamed to `point_25d_legacy()` (and a new
        function with the name `point_25d()` was introduced). See notes for
        further details.

    Parameters
    ----------
    omega : float
        Angular frequency of point source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of virtual point source.
    xref : (3,) array_like, optional
        Reference point for synthesized sound field.
    c : float, optional
        Speed of sound.
    omalias: float, optional
        Angular frequency where spatial aliasing becomes prominent.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    `point_25d_legacy()` derives 2.5D WFS from the 2D
    Neumann-Rayleigh integral (i.e. the approach by Rabenstein & Spors), cf.
    :cite:`Spors2008`.

    .. math::

        D(\x_0,\w) = \sqrt{\i\wc |\x_\text{ref}-\x_0|}
            \frac{\scalarprod{\x_0-\x_\text{s}}{\n_0}}
            {|\x_0-\x_\text{s}|^\frac{3}{2}}
            \e{-\i\wc |\x_0-\x_\text{s}|}

    The theoretical link of `point_25d()` and `point_25d_legacy()` was
    introduced as *unified WFS framework* in :cite:`Firtha2017`.
    Also cf. Eq. (2.145)-(2.147) :cite:`Schultz2016`.

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.point_25d_legacy(
            omega, array.x, array.n, xs)
        normalize_gain = np.linalg.norm(xs)
        plot(normalize_gain * d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    xref = util.asarray_1d(xref)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = (
        preeq_25d(omega, omalias, c) *
        np.sqrt(np.linalg.norm(xref - x0)) * inner1d(ds, n0) /
        r ** (3 / 2) * np.exp(-1j * k * r))
    selection = util.source_selection_point(n0, x0, xs)
    return d, selection, secondary_source_point(omega, c)


def _plane(omega, x0, n0, n=[0, 1, 0], *, c=None):
    r"""Driving function for 2/3-dimensional WFS for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of plane wave.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    n : (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    Eq.(17) from :cite:`Spors2008`:

    .. math::

        D(\x_0,\w) = \i\wc \scalarprod{\n}{\n_0}
            \e{-\i\wc\scalarprod{\n}{\x_0}}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.plane_3d(
            omega, array.x, array.n, npw)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    n = util.normalize_vector(n)
    k = util.wavenumber(omega, c)
    d = 2j * k * np.inner(n, n0) * np.exp(-1j * k * np.inner(n, x0))
    selection = util.source_selection_plane(n0, n)
    return d, selection, secondary_source_point(omega, c)


plane_2d = _plane


def plane_25d(omega, x0, n0, n=[0, 1, 0], *, xref=[0, 0, 0], c=None,
              omalias=None):
    r"""Driving function for 2.5-dimensional WFS for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of plane wave.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    n : (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    xref : (3,) array_like, optional
        Reference point for synthesized sound field.
    c : float, optional
        Speed of sound.
    omalias: float, optional
        Angular frequency where spatial aliasing becomes prominent.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D_\text{2.5D}(\x_0,\w) = \sqrt{\i\wc |\x_\text{ref}-\x_0|}
            \scalarprod{\n}{\n_0}
            \e{-\i\wc \scalarprod{\n}{\x_0}}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.plane_25d(
            omega, array.x, array.n, npw)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    n = util.normalize_vector(n)
    xref = util.asarray_1d(xref)
    k = util.wavenumber(omega, c)
    d = (
        preeq_25d(omega, omalias, c) *
        np.sqrt(8*np.pi * np.linalg.norm(xref - x0, axis=-1)) *
        np.inner(n, n0) * np.exp(-1j * k * np.inner(n, x0)))
    selection = util.source_selection_plane(n0, n)
    return d, selection, secondary_source_point(omega, c)


plane_3d = _plane


def _focused(omega, x0, n0, xs, ns, *, c=None):
    r"""Driving function for 2/3-dimensional WFS for a focused source.

    Parameters
    ----------
    omega : float
        Angular frequency of focused source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of focused source.
    ns :  (3,) array_like
        Direction of focused source.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\x_0,\w) = \i\wc \frac{\scalarprod{\x_0-\x_\text{s}}{\n_0}}
            {|\x_0-\x_\text{s}|^\frac{3}{2}}
            \e{\i\wc |\x_0-\x_\text{s}|}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.focused_3d(
            omega, array.x, array.n, xs_focused, ns_focused)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = 1j * k * inner1d(ds, n0) / r ** (3 / 2) * np.exp(1j * k * r)
    selection = util.source_selection_focused(ns, x0, xs)
    return d, selection, secondary_source_point(omega, c)


focused_2d = _focused


def focused_25d(omega, x0, n0, xs, ns, *, xref=[0, 0, 0], c=None,
                omalias=None):
    r"""Driving function for 2.5-dimensional WFS for a focused source.

    Parameters
    ----------
    omega : float
        Angular frequency of focused source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of focused source.
    ns :  (3,) array_like
        Direction of focused source.
    xref : (3,) array_like, optional
        Reference point for synthesized sound field.
    c : float, optional
        Speed of sound.
    omalias: float, optional
        Angular frequency where spatial aliasing becomes prominent.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\x_0,\w) = \sqrt{\i\wc |\x_\text{ref}-\x_0|}
            \frac{\scalarprod{\x_0-\x_\text{s}}{\n_0}}
            {|\x_0-\x_\text{s}|^\frac{3}{2}}
            \e{\i\wc |\x_0-\x_\text{s}|}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.focused_25d(
            omega, array.x, array.n, xs_focused, ns_focused)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    xref = util.asarray_1d(xref)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = (
        preeq_25d(omega, omalias, c) *
        np.sqrt(np.linalg.norm(xref - x0)) * inner1d(ds, n0) /
        r ** (3 / 2) * np.exp(1j * k * r))
    selection = util.source_selection_focused(ns, x0, xs)
    return d, selection, secondary_source_point(omega, c)


focused_3d = _focused


def preeq_25d(omega, omalias, c):
    r"""Pre-equalization filter for 2.5-dimensional WFS.

    Parameters
    ----------
    omega : float
        Angular frequency.
    omalias: float
        Angular frequency where spatial aliasing becomes prominent.
    c : float
        Speed of sound.

    Returns
    -------
    float
        Complex weight for given angular frequency.

    Notes
    -----
    .. math::

        H(\w) = \begin{cases}
            \sqrt{\i \wc} & \text{for } \w \leq \w_\text{alias} \\
            \sqrt{\i \frac{\w_\text{alias}}{c}} &
            \text{for } \w > \w_\text{alias}
            \end{cases}

    """
    if omalias is None:
        return np.sqrt(1j * util.wavenumber(omega, c))
    else:
        if omega <= omalias:
            return np.sqrt(1j * util.wavenumber(omega, c))
        else:
            return np.sqrt(1j * util.wavenumber(omalias, c))


def plane_3d_delay(omega, x0, n0, n=[0, 1, 0], *, c=None):
    r"""Delay-only driving function for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of plane wave.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    n : (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    .. math::

        D(\x_0,\w) = \e{-\i\wc\scalarprod{\n}{\x_0}}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.wfs.plane_3d_delay(
            omega, array.x, array.n, npw)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n = util.normalize_vector(n)
    k = util.wavenumber(omega, c)
    d = np.exp(-1j * k * np.inner(n, x0))
    selection = util.source_selection_plane(n0, n)
    return d, selection, secondary_source_point(omega, c)


def soundfigure_3d(omega, x0, n0, figure, npw=[0, 0, 1], *, c=None):
    """Compute driving function for a 2D sound figure.

    Based on
    [Helwani et al., The Synthesis of Sound Figures, MSSP, 2013]

    """
    x0 = np.asarray(x0)
    n0 = np.asarray(n0)
    k = util.wavenumber(omega, c)
    nx, ny = figure.shape

    # 2D spatial DFT of image
    figure = np.fft.fftshift(figure, axes=(0, 1))  # sign of spatial DFT
    figure = np.fft.fft2(figure)
    # wavenumbers
    kx = np.fft.fftfreq(nx, 1./nx)
    ky = np.fft.fftfreq(ny, 1./ny)
    # shift spectrum due to desired plane wave
    figure = np.roll(figure, int(k*npw[0]), axis=0)
    figure = np.roll(figure, int(k*npw[1]), axis=1)
    # search and iterate over propagating plane wave components
    kxx, kyy = np.meshgrid(kx, ky, sparse=True)
    rho = np.sqrt((kxx) ** 2 + (kyy) ** 2)
    d = 0
    for n in range(nx):
        for m in range(ny):
            if(rho[n, m] < k):
                # dispertion relation
                kz = np.sqrt(k**2 - rho[n, m]**2)
                # normal vector of plane wave
                npw = 1/k * np.asarray([kx[n], ky[m], kz])
                npw = npw / np.linalg.norm(npw)
                # driving function of plane wave with positive kz
                d_component, selection, secondary_source = plane_3d(
                    omega, x0, n0, npw, c=c)
                d += selection * figure[n, m] * d_component

    return d, util.source_selection_all(len(d)), secondary_source



"""Submodules for broadband sound fields.

.. autosummary::
    :toctree:

    source

    wfs
    nfchoa

"""
from .. import array as _array
import numpy as np
from . import source
from .. import util as _util


def synthesize(signals, weights, ssd, secondary_source_function, **kwargs):
    """Compute sound field for an array of secondary sources.

    Parameters
    ----------
    signals : (N, C) array_like + float
        Driving signals consisting of audio data (C channels) and a
        sampling rate (in Hertz).
        A `DelayedSignal` object can also be used.
    weights : (C,) array_like
        Additional weights applied during integration, e.g. source
        selection and tapering.
    ssd : sequence of between 1 and 3 array_like objects
        Positions (shape ``(C, 3)``), normal vectors (shape ``(C, 3)``)
        and weights (shape ``(C,)``) of secondary sources.
        A `SecondarySourceDistribution` can also be used.
    secondary_source_function : callable
        A function that generates the sound field of a secondary source.
        This signature is expected::

            secondary_source_function(
                position, normal_vector, weight, driving_signal,
                **kwargs) -> numpy.ndarray

    **kwargs
        All keyword arguments are forwarded to *secondary_source_function*.
        This is typically used to pass the *observation_time* and *grid*
        arguments.

    Returns
    -------
    numpy.ndarray
        Sound pressure at grid positions.

    """
    ssd = _array.as_secondary_source_distribution(ssd)
    data, samplerate, signal_offset = _util.as_delayed_signal(signals)
    weights = _util.asarray_1d(weights)
    channels = data.T
    if not (len(ssd.x) == len(ssd.n) == len(ssd.a) == len(channels) ==
            len(weights)):
        raise ValueError("Length mismatch")
    p = 0
    for x, n, a, channel, weight in zip(ssd.x, ssd.n, ssd.a,
                                        channels, weights):
        if weight != 0:
            signal = channel, samplerate, signal_offset
            p += a * weight * secondary_source_function(x, n, signal, **kwargs)
    return p


def apply_delays(signal, delays):
    """Apply delays for every channel.

    Parameters
    ----------
    signal : (N,) array_like + float
        Excitation signal consisting of (mono) audio data and a sampling
        rate (in Hertz).  A `DelayedSignal` object can also be used.
    delays : (C,) array_like
        Delay in seconds for each channel (C), negative values allowed.

    Returns
    -------
    `DelayedSignal`
        A tuple containing the delayed signals (in a `numpy.ndarray`
        with shape ``(N, C)``), followed by the sampling rate (in Hertz)
        and a (possibly negative) time offset (in seconds).

    """
    data, samplerate, initial_offset = _util.as_delayed_signal(signal)
    data = _util.asarray_1d(data)
    delays = _util.asarray_1d(delays)
    delays += initial_offset

    delays_samples = np.rint(samplerate * delays).astype(int)
    offset_samples = delays_samples.min()
    delays_samples -= offset_samples
    out = np.zeros((delays_samples.max() + len(data), len(delays_samples)))
    for column, row in enumerate(delays_samples):
        out[row:row + len(data), column] = data
    return _util.DelayedSignal(out, samplerate, offset_samples / samplerate)


def secondary_source_point(c):
    """Create a point source for use in `sfs.td.synthesize()`."""

    def secondary_source(position, _, signal, observation_time, grid):
        return source.point(position, signal, observation_time, grid, c=c)

    return secondary_source


from . import nfchoa
from . import wfs



"""Compute positions of various secondary source distributions.

.. plot::
    :context: reset

    import sfs
    import matplotlib.pyplot as plt
    plt.rcParams['figure.figsize'] = 8, 4.5  # inch
    plt.rcParams['axes.grid'] = True

"""
from collections import namedtuple
import numpy as np
from . import util


class SecondarySourceDistribution(namedtuple('SecondarySourceDistribution',
                                             'x n a')):
    """Named tuple returned by array functions.

    See `collections.namedtuple`.

    Attributes
    ----------
    x : (N, 3) numpy.ndarray
        Positions of secondary sources
    n : (N, 3) numpy.ndarray
        Orientations (normal vectors) of secondary sources
    a : (N,) numpy.ndarray
        Weights of secondary sources

    """

    __slots__ = ()

    def __repr__(self):
        return 'SecondarySourceDistribution(\n' + ',\n'.join(
            '    {}={}'.format(name, repr(data).replace('\n', '\n      '))
            for name, data in zip('xna', self)) + ')'

    def take(self, indices):
        """Return a sub-array given by *indices*."""
        return SecondarySourceDistribution(
            self.x[indices], self.n[indices], self.a[indices])


def as_secondary_source_distribution(arg, **kwargs):
    r"""Create a `SecondarySourceDistribution`.

    Parameters
    ----------
    arg : sequence of between 1 and 3 array_like objects
        All elements are converted to NumPy arrays.
        If only 1 element is given, all normal vectors are set to *NaN*.
        If only 1 or 2 elements are given, all weights are set to ``1.0``.
    **kwargs
        All keyword arguments are forwarded to :func:`numpy.asarray`.

    Returns
    -------
    `SecondarySourceDistribution`
        A named tuple consisting of three `numpy.ndarray`\s containing
        positions, normal vectors and weights.

    """
    if len(arg) == 3:
        x, n, a = arg
    elif len(arg) == 2:
        x, n = arg
        a = 1.0
    elif len(arg) == 1:
        x, = arg
        n = np.nan, np.nan, np.nan
        a = 1.0
    else:
        raise TypeError('Between 1 and 3 elements are required')
    x = util.asarray_of_rows(x, **kwargs)
    n = util.asarray_of_rows(n, **kwargs)
    if len(n) == 1:
        n = np.tile(n, (len(x), 1))
    a = util.asarray_1d(a, **kwargs)
    if len(a) == 1:
        a = np.tile(a, len(x))
    return SecondarySourceDistribution(x, n, a)


def linear(N, spacing, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return linear, equidistantly sampled secondary source distribution.

    Parameters
    ----------
    N : int
        Number of secondary sources.
    spacing : float
        Distance (in metres) between secondary sources.
    center : (3,) array_like, optional
        Coordinates of array center.
    orientation : (3,) array_like, optional
        Orientation of the array.  By default, the loudspeakers have
        their main axis pointing into positive x-direction.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.linear(16, 0.2, orientation=[0, -1, 0])
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    return _linear_helper(np.arange(N) * spacing, center, orientation)


def linear_diff(distances, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return linear secondary source distribution from a list of distances.

    Parameters
    ----------
    distances : (N-1,) array_like
        Sequence of secondary sources distances in metres.
    center, orientation
        See `linear()`.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.linear_diff(4 * [0.3] + 6 * [0.15] + 4 * [0.3],
                                           orientation=[0, -1, 0])
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    distances = util.asarray_1d(distances)
    ycoordinates = np.concatenate(([0], np.cumsum(distances)))
    return _linear_helper(ycoordinates, center, orientation)


def linear_random(N, min_spacing, max_spacing, *, center=[0, 0, 0],
                  orientation=[1, 0, 0], seed=None):
    """Return randomly sampled linear array.

    Parameters
    ----------
    N : int
        Number of secondary sources.
    min_spacing, max_spacing : float
        Minimal and maximal distance (in metres) between secondary
        sources.
    center, orientation
        See `linear()`.
    seed : {None, int, array_like}, optional
        Random seed.  See `numpy.random.RandomState`.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.linear_random(
            N=12,
            min_spacing=0.15, max_spacing=0.4,
            orientation=[0, -1, 0])
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    r = np.random.RandomState(seed)
    distances = r.uniform(min_spacing, max_spacing, size=N-1)
    return linear_diff(distances, center=center, orientation=orientation)


def circular(N, R, *, center=[0, 0, 0]):
    """Return circular secondary source distribution parallel to the xy-plane.

    Parameters
    ----------
    N : int
        Number of secondary sources.
    R : float
        Radius in metres.
    center
        See `linear()`.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.circular(16, 1)
        sfs.plot2d.loudspeakers(x0, n0, a0, size=0.2, show_numbers=True)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    center = util.asarray_1d(center)
    alpha = np.linspace(0, 2 * np.pi, N, endpoint=False)
    positions = np.zeros((N, len(center)))
    positions[:, 0] = R * np.cos(alpha)
    positions[:, 1] = R * np.sin(alpha)
    positions += center
    normals = np.zeros_like(positions)
    normals[:, 0] = np.cos(alpha + np.pi)
    normals[:, 1] = np.sin(alpha + np.pi)
    weights = np.ones(N) * 2 * np.pi * R / N
    return SecondarySourceDistribution(positions, normals, weights)


def rectangular(N, spacing, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return rectangular secondary source distribution.

    Parameters
    ----------
    N : int or pair of int
        Number of secondary sources on each side of the rectangle.
        If a pair of numbers is given, the first one specifies the first
        and third segment, the second number specifies the second and
        fourth segment.
    spacing : float
        Distance (in metres) between secondary sources.
    center, orientation
        See `linear()`.  The *orientation* corresponds to the first
        linear segment.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.rectangular((4, 8), 0.2)
        sfs.plot2d.loudspeakers(x0, n0, a0, show_numbers=True)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    N1, N2 = (N, N) if np.isscalar(N) else N
    offset1 = spacing * (N2 - 1) / 2 + spacing / np.sqrt(2)
    offset2 = spacing * (N1 - 1) / 2 + spacing / np.sqrt(2)
    positions, normals, weights = concatenate(
        # left
        linear(N1, spacing, center=[-offset1, 0, 0], orientation=[1, 0, 0]),
        # upper
        linear(N2, spacing, center=[0, offset2, 0], orientation=[0, -1, 0]),
        # right
        linear(N1, spacing, center=[offset1, 0, 0], orientation=[-1, 0, 0]),
        # lower
        linear(N2, spacing, center=[0, -offset2, 0], orientation=[0, 1, 0]),
    )
    positions, normals = _rotate_array(positions, normals,
                                       [1, 0, 0], orientation)
    positions += center
    return SecondarySourceDistribution(positions, normals, weights)


def rounded_edge(Nxy, Nr, dx, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return SSD along the xy-axis with rounded edge at the origin.

    Parameters
    ----------
    Nxy : int
        Number of secondary sources along x- and y-axis.
    Nr : int
        Number of secondary sources in rounded edge.  Radius of edge is
        adjusted to equdistant sampling along entire array.
    center : (3,) array_like, optional
        Position of edge.
    orientation : (3,) array_like, optional
        Normal vector of array.  Default orientation is along xy-axis.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.rounded_edge(8, 5, 0.2)
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    # radius of rounded edge
    Nr += 1
    R = 2/np.pi * Nr * dx

    # array along y-axis
    x00, n00, a00 = linear(Nxy, dx, center=[0, Nxy//2*dx+dx/2+R, 0])
    x00 = np.flipud(x00)
    positions = x00
    directions = n00
    weights = a00

    # round part
    x00 = np.zeros((Nr, 3))
    n00 = np.zeros((Nr, 3))
    a00 = np.zeros(Nr)
    for n in range(0, Nr):
        alpha = np.pi/2 * n/Nr
        x00[n, 0] = R * (1-np.cos(alpha))
        x00[n, 1] = R * (1-np.sin(alpha))
        n00[n, 0] = np.cos(alpha)
        n00[n, 1] = np.sin(alpha)
        a00[n] = dx
    positions = np.concatenate((positions, x00))
    directions = np.concatenate((directions, n00))
    weights = np.concatenate((weights, a00))

    # array along x-axis
    x00, n00, a00 = linear(Nxy, dx, center=[Nxy//2*dx-dx/2+R, 0, 0],
                           orientation=[0, 1, 0])
    x00 = np.flipud(x00)
    positions = np.concatenate((positions, x00))
    directions = np.concatenate((directions, n00))
    weights = np.concatenate((weights, a00))

    # rotate array
    positions, directions = _rotate_array(positions, directions,
                                          [1, 0, 0], orientation)
    # shift array to desired position
    positions += center
    return SecondarySourceDistribution(positions, directions, weights)


def edge(Nxy, dx, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return SSD along the xy-axis with sharp edge at the origin.

    Parameters
    ----------
    Nxy : int
        Number of secondary sources along x- and y-axis.
    center : (3,) array_like, optional
        Position of edge.
    orientation : (3,) array_like, optional
        Normal vector of array.  Default orientation is along xy-axis.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.edge(8, 0.2)
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    # array along y-axis
    x00, n00, a00 = linear(Nxy, dx, center=[0, Nxy//2*dx+dx/2, 0])
    x00 = np.flipud(x00)
    positions = x00
    directions = n00
    weights = a00

    # array along x-axis
    x00, n00, a00 = linear(Nxy, dx, center=[Nxy//2*dx-dx/2, 0, 0],
                           orientation=[0, 1, 0])
    x00 = np.flipud(x00)
    positions = np.concatenate((positions, x00))
    directions = np.concatenate((directions, n00))
    weights = np.concatenate((weights, a00))

    # rotate array
    positions, directions = _rotate_array(positions, directions,
                                          [1, 0, 0], orientation)
    # shift array to desired position
    positions += center
    return SecondarySourceDistribution(positions, directions, weights)


def planar(N, spacing, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return planar secondary source distribtion.

    Parameters
    ----------
    N : int or pair of int
        Number of secondary sources along each edge.
        If a pair of numbers is given, the first one specifies the
        number on the horizontal edge, the second one specifies the
        number on the vertical edge.
    spacing : float
        Distance (in metres) between secondary sources.
    center, orientation
        See `linear()`.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.planar(
            (4,3), 0.5, orientation=[0, 0, 1])  # 4 sources along y, 3 sources along x
        x0, n0, a0 = sfs.array.planar(
            (4,3), 0.5, orientation=[1, 0, 0])  # 4 sources along y, 3 sources along z

        x0, n0, a0 = sfs.array.planar(
            (4,3), 0.5, orientation=[0, 1, 0])  # 4 sources along x, 3 sources along z
        sfs.plot2d.loudspeakers(x0, n0, a0)  # plot the last ssd in 2D
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')


    """
    N1, N2 = (N, N) if np.isscalar(N) else N
    zcoordinates = np.arange(N2) * spacing
    zcoordinates -= np.mean(zcoordinates[[0, -1]])  # move center to origin
    subarrays = [linear(N1, spacing, center=[0, 0, z]) for z in zcoordinates]
    positions, normals, weights = concatenate(*subarrays)
    weights *= spacing
    positions, normals = _rotate_array(positions, normals,
                                       [1, 0, 0], orientation)
    positions += center
    return SecondarySourceDistribution(positions, normals, weights)


def cube(N, spacing, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Return cube-shaped secondary source distribtion.

    Parameters
    ----------
    N : int or triple of int
        Number of secondary sources along each edge.  If a triple of
        numbers is given, the first two specify the edges like in
        `rectangular()`, the last one specifies the vertical edge.
    spacing : float
        Distance (in metres) between secondary sources.
    center, orientation
        See `linear()`.  The *orientation* corresponds to the first
        planar segment.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.cube(
            N=2, spacing=0.5,
            center=[0, 0, 0], orientation=[1, 0, 0])
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')
        plt.title('view onto xy-plane')

    """
    N1, N2, N3 = (N, N, N) if np.isscalar(N) else N
    d = spacing
    offset1 = d * (N2 - 1) / 2 + d / np.sqrt(2)
    offset2 = d * (N1 - 1) / 2 + d / np.sqrt(2)
    offset3 = d * (N3 - 1) / 2 + d / np.sqrt(2)
    positions, directions, weights = concatenate(
        # west
        planar((N1, N3), d, center=[-offset1, 0, 0], orientation=[1, 0, 0]),
        # north
        planar((N2, N3), d, center=[0, offset2, 0], orientation=[0, -1, 0]),
        # east
        planar((N1, N3), d, center=[offset1, 0, 0], orientation=[-1, 0, 0]),
        # south
        planar((N2, N3), d, center=[0, -offset2, 0], orientation=[0, 1, 0]),
        # bottom
        planar((N2, N1), d, center=[0, 0, -offset3], orientation=[0, 0, 1]),
        # top
        planar((N2, N1), d, center=[0, 0, offset3], orientation=[0, 0, -1]),
    )
    positions, directions = _rotate_array(positions, directions,
                                          [1, 0, 0], orientation)
    positions += center
    return SecondarySourceDistribution(positions, directions, weights)


def sphere_load(file, radius, *, center=[0, 0, 0]):
    """Load spherical secondary source distribution from file.

    ASCII Format (see MATLAB SFS Toolbox) with 4 numbers (3 for the cartesian
    position vector, 1 for the integration weight) per secondary source located
    on the unit circle which is resized by the given radius and shifted to the
    given center.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    content of ``example_array_6LS_3D.txt``::

        1 0 0 1
        -1 0 0 1
        0 1 0 1
        0 -1 0 1
        0 0 1 1
        0 0 -1 1

    corresponds to the `3-dimensional 6-point spherical 3-design
    <http://neilsloane.com/sphdesigns/dim3/des.3.6.3.txt>`_.

    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.sphere_load(
            '../data/arrays/example_array_6LS_3D.txt',
            radius=2,
            center=[0, 0, 0])
        sfs.plot2d.loudspeakers(x0, n0, a0, size=0.25)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')
        plt.title('view onto xy-plane')

    """
    data = np.loadtxt(file)
    positions, weights = data[:, :3], data[:, 3]
    normals = -positions
    positions *= radius
    positions += center
    return SecondarySourceDistribution(positions, normals, weights)


def load(file, *, center=[0, 0, 0], orientation=[1, 0, 0]):
    """Load secondary source distribution from file.

    Comma Separated Values (CSV) format with 7 values
    (3 for the cartesian position vector, 3 for the cartesian inward normal
    vector, 1 for the integration weight) per secondary source.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights of secondary sources.

    Examples
    --------
    content of ``example_array_4LS_2D.csv``::

        1,0,0,-1,0,0,1
        0,1,0,0,-1,0,1
        -1,0,0,1,0,0,1
        0,-1,0,0,1,0,1

    corresponds to 4 sources at 1, j, -1, -j in the complex plane. This setup
    is typically used for Quadraphonic audio reproduction.

    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.load('../data/arrays/example_array_4LS_2D.csv')
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    .. plot::
        :context: close-figs

        x0, n0, a0 = sfs.array.load(
            '../data/arrays/wfs_university_rostock_2018.csv')
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')
        plt.title('top view of 64 channel WFS system at university of Rostock')

    """
    data = np.loadtxt(file, delimiter=',')
    positions, normals, weights = data[:, :3], data[:, 3:6], data[:, 6]
    positions, normals = _rotate_array(positions, normals,
                                       [1, 0, 0], orientation)
    positions += center
    return SecondarySourceDistribution(positions, normals, weights)


def weights_midpoint(positions, *, closed):
    """Calculate loudspeaker weights for a simply connected array.

    The weights are calculated according to the midpoint rule.


    Parameters
    ----------
    positions : (N, 3) array_like
        Sequence of secondary source positions.

        .. note:: The loudspeaker positions have to be ordered along the
                  contour.

    closed : bool
        ``True`` if the loudspeaker contour is closed.

    Returns
    -------
    (N,) numpy.ndarray
        Weights of secondary sources.

    Examples
    --------
    >>> import sfs
    >>> x0, n0, a0 = sfs.array.circular(2**5, 1)
    >>> a = sfs.array.weights_midpoint(x0, closed=True)
    >>> max(abs(a0-a))
    0.0003152601902411123

    """
    positions = util.asarray_of_rows(positions)
    if closed:
        before, after = -1, 0  # cyclic
    else:
        before, after = 1, -2  # mirrored
    positions = np.row_stack((positions[before], positions, positions[after]))
    distances = np.linalg.norm(np.diff(positions, axis=0), axis=1)
    return (distances[:-1] + distances[1:]) / 2


def _rotate_array(positions, normals, n1, n2):
    """Rotate secondary sources from n1 to n2."""
    R = util.rotation_matrix(n1, n2)
    positions = np.inner(positions, R)
    normals = np.inner(normals, R)
    return positions, normals


def _linear_helper(ycoordinates, center, orientation):
    """Create a full linear array from an array of y-coordinates."""
    center = util.asarray_1d(center)
    N = len(ycoordinates)
    positions = np.zeros((N, 3))
    positions[:, 1] = ycoordinates - np.mean(ycoordinates[[0, -1]])
    positions, normals = _rotate_array(positions, [1, 0, 0],
                                       [1, 0, 0], orientation)
    positions += center
    normals = np.tile(normals, (N, 1))
    weights = weights_midpoint(positions, closed=False)
    return SecondarySourceDistribution(positions, normals, weights)


def concatenate(*arrays):
    """Concatenate `SecondarySourceDistribution` objects.

    Returns
    -------
    `SecondarySourceDistribution`
        Positions, orientations and weights
        of the concatenated secondary sources.

    Examples
    --------
    .. plot::
        :context: close-figs

        ssd1 = sfs.array.edge(10, 0.2)
        ssd2 = sfs.array.edge(20, 0.1, center=[2, 2, 0], orientation=[-1, 0, 0])
        x0, n0, a0 = sfs.array.concatenate(ssd1, ssd2)
        sfs.plot2d.loudspeakers(x0, n0, a0)
        plt.axis('equal')
        plt.xlabel('x / m')
        plt.ylabel('y / m')

    """
    return SecondarySourceDistribution._make(np.concatenate(i)
                                             for i in zip(*arrays))



"""Compute NFC-HOA driving functions.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import matplotlib.pyplot as plt
    import numpy as np
    import sfs
    from scipy.signal import unit_impulse

    # Plane wave
    npw = sfs.util.direction_vector(np.radians(-45))

    # Point source
    xs = -1.5, 1.5, 0
    rs = np.linalg.norm(xs)  # distance from origin
    ts = rs / sfs.default.c  # time-of-arrival at origin

    # Impulsive excitation
    fs = 44100
    signal = unit_impulse(512), fs

    # Circular loudspeaker array
    N = 32  # number of loudspeakers
    R = 1.5  # radius
    array = sfs.array.circular(N, R)

    grid = sfs.util.xyz_grid([-2, 2], [-2, 2], 0, spacing=0.02)

    def plot(d, selection, secondary_source, t=0):
        p = sfs.td.synthesize(d, selection, array, secondary_source, grid=grid,
                              observation_time=t)
        sfs.plot2d.level(p, grid)
        sfs.plot2d.loudspeakers(array.x, array.n, selection * array.a, size=0.15)

"""
import numpy as np
from scipy.signal import besselap, sosfilt, zpk2sos
from scipy.special import eval_legendre as legendre
from .. import default
from .. import util
from . import secondary_source_point


def matchedz_zpk(s_zeros, s_poles, s_gain, fs):
    """Matched-z transform of poles and zeros.

    Parameters
    ----------
    s_zeros : array_like
        Zeros in the Laplace domain.
    s_poles : array_like
        Poles in the Laplace domain.
    s_gain : float
        System gain in the Laplace domain.
    fs : int
        Sampling frequency in Hertz.

    Returns
    -------
    z_zeros : numpy.ndarray
        Zeros in the z-domain.
    z_poles : numpy.ndarray
        Poles in the z-domain.
    z_gain : float
        System gain in the z-domain.

    See Also
    --------
    :func:`scipy.signal.bilinear_zpk`

    """
    z_zeros = np.exp(s_zeros / fs)
    z_poles = np.exp(s_poles / fs)
    omega = 1j * np.pi * fs
    s_gain *= np.prod((omega - s_zeros) / (omega - s_poles)
                      * (-1 - z_poles) / (-1 - z_zeros))
    return z_zeros, z_poles, np.real(s_gain)


def plane_25d(x0, r0, npw, fs, max_order=None, c=None, s2z=matchedz_zpk):
    r"""Virtual plane wave by 2.5-dimensional NFC-HOA.

    .. math::

        D(\phi_0, s) =
        2\e{\frac{s}{c}r_0}
        \sum_{m=-M}^{M}
        (-1)^m
        \Big(\frac{s}{s-\frac{c}{r_0}\sigma_0}\Big)^\mu
        \prod_{l=1}^{\nu}
        \frac{s^2}{(s-\frac{c}{r_0}\sigma_l)^2+(\frac{c}{r_0}\omega_l)^2}
        \e{\i m(\phi_0 - \phi_\text{pw})}

    The driving function is represented in the Laplace domain,
    from which the recursive filters are designed.
    :math:`\sigma_l + \i\omega_l` denotes the complex roots of
    the reverse Bessel polynomial.
    The number of second-order sections is
    :math:`\nu = \big\lfloor\tfrac{|m|}{2}\big\rfloor`,
    whereas the number of first-order section :math:`\mu` is either 0 or 1
    for even and odd :math:`|m|`, respectively.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of the circular secondary source distribution.
    npw : (3,) array_like
        Unit vector (propagation direction) of plane wave.
    fs : int
        Sampling frequency in Hertz.
    max_order : int, optional
        Ambisonics order.
    c : float, optional
        Speed of sound in m/s.
    s2z : callable, optional
        Function transforming s-domain poles and zeros into z-domain,
        e.g. :func:`matchedz_zpk`, :func:`scipy.signal.bilinear_zpk`.

    Returns
    -------
    delay : float
        Overall delay in seconds.
    weight : float
        Overall weight.
    sos : list of numpy.ndarray
        Second-order section filters :func:`scipy.signal.sosfilt`.
    phaseshift : (N,) numpy.ndarray
        Phase shift in radians.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    Examples
    --------
    .. plot::
        :context: close-figs

        delay, weight, sos, phaseshift, selection, secondary_source = \
            sfs.td.nfchoa.plane_25d(array.x, R, npw, fs)
        d = sfs.td.nfchoa.driving_signals_25d(
                delay, weight, sos, phaseshift, signal)
        plot(d, selection, secondary_source)

    """
    if max_order is None:
        max_order = util.max_order_circular_harmonics(len(x0))
    if c is None:
        c = default.c

    x0 = util.asarray_of_rows(x0)
    npw = util.asarray_1d(npw)
    phi0, _, _ = util.cart2sph(*x0.T)
    phipw, _, _ = util.cart2sph(*npw)
    phaseshift = phi0 - phipw + np.pi

    delay = -r0 / c
    weight = 2
    sos = []
    for m in range(max_order + 1):
        _, p, _ = besselap(m, norm='delay')
        s_zeros = np.zeros(m)
        s_poles = c / r0 * p
        s_gain = 1
        z_zeros, z_poles, z_gain = s2z(s_zeros, s_poles, s_gain, fs)
        sos.append(zpk2sos(z_zeros, z_poles, z_gain, pairing='nearest'))
    selection = util.source_selection_all(len(x0))
    return delay, weight, sos, phaseshift, selection, secondary_source_point(c)


def point_25d(x0, r0, xs, fs, max_order=None, c=None, s2z=matchedz_zpk):
    r"""Virtual Point source by 2.5-dimensional NFC-HOA.

    .. math::

        D(\phi_0, s) =
        \frac{1}{2\pi r_\text{s}}
        \e{\frac{s}{c}(r_0-r_\text{s})}
        \sum_{m=-M}^{M}
        \Big(\frac{s-\frac{c}{r_\text{s}}\sigma_0}{s-\frac{c}{r_0}\sigma_0}\Big)^\mu
        \prod_{l=1}^{\nu}
        \frac{(s-\frac{c}{r_\text{s}}\sigma_l)^2-(\frac{c}{r_\text{s}}\omega_l)^2}
        {(s-\frac{c}{r_0}\sigma_l)^2+(\frac{c}{r_0}\omega_l)^2}
        \e{\i m(\phi_0 - \phi_\text{s})}

    The driving function is represented in the Laplace domain,
    from which the recursive filters are designed.
    :math:`\sigma_l + \i\omega_l` denotes the complex roots of
    the reverse Bessel polynomial.
    The number of second-order sections is
    :math:`\nu = \big\lfloor\tfrac{|m|}{2}\big\rfloor`,
    whereas the number of first-order section :math:`\mu` is either 0 or 1
    for even and odd :math:`|m|`, respectively.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of the circular secondary source distribution.
    xs : (3,) array_like
        Virtual source position.
    fs : int
        Sampling frequency in Hertz.
    max_order : int, optional
        Ambisonics order.
    c : float, optional
        Speed of sound in m/s.
    s2z : callable, optional
        Function transforming s-domain poles and zeros into z-domain,
        e.g. :func:`matchedz_zpk`, :func:`scipy.signal.bilinear_zpk`.

    Returns
    -------
    delay : float
        Overall delay in seconds.
    weight : float
        Overall weight.
    sos : list of numpy.ndarray
        Second-order section filters :func:`scipy.signal.sosfilt`.
    phaseshift : (N,) numpy.ndarray
        Phase shift in radians.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    Examples
    --------
    .. plot::
        :context: close-figs

        delay, weight, sos, phaseshift, selection, secondary_source = \
            sfs.td.nfchoa.point_25d(array.x, R, xs, fs)
        d = sfs.td.nfchoa.driving_signals_25d(
                delay, weight, sos, phaseshift, signal)
        plot(d, selection, secondary_source, t=ts)

    """
    if max_order is None:
        max_order = util.max_order_circular_harmonics(len(x0))
    if c is None:
        c = default.c

    x0 = util.asarray_of_rows(x0)
    xs = util.asarray_1d(xs)
    phi0, _, _ = util.cart2sph(*x0.T)
    phis, _, rs = util.cart2sph(*xs)
    phaseshift = phi0 - phis

    delay = (rs - r0) / c
    weight = 1 / 2 / np.pi / rs
    sos = []
    for m in range(max_order + 1):
        _, p, _ = besselap(m, norm='delay')
        s_zeros = c / rs * p
        s_poles = c / r0 * p
        s_gain = 1
        z_zeros, z_poles, z_gain = s2z(s_zeros, s_poles, s_gain, fs)
        sos.append(zpk2sos(z_zeros, z_poles, z_gain, pairing='nearest'))
    selection = util.source_selection_all(len(x0))
    return delay, weight, sos, phaseshift, selection, secondary_source_point(c)


def plane_3d(x0, r0, npw, fs, max_order=None, c=None, s2z=matchedz_zpk):
    r"""Virtual plane wave by 3-dimensional NFC-HOA.

    .. math::

        D(\phi_0, s) =
        \frac{\e{\frac{s}{c}r_0}}{r_0}
        \sum_{n=0}^{N}
        (-1)^n (2n+1) P_{n}(\cos\Theta)
        \Big(\frac{s}{s-\frac{c}{r_0}\sigma_0}\Big)^\mu
        \prod_{l=1}^{\nu}
        \frac{s^2}{(s-\frac{c}{r_0}\sigma_l)^2+(\frac{c}{r_0}\omega_l)^2}

    The driving function is represented in the Laplace domain,
    from which the recursive filters are designed.
    :math:`\sigma_l + \i\omega_l` denotes the complex roots of
    the reverse Bessel polynomial.
    The number of second-order sections is
    :math:`\nu = \big\lfloor\tfrac{|m|}{2}\big\rfloor`,
    whereas the number of first-order section :math:`\mu` is either 0 or 1
    for even and odd :math:`|m|`, respectively.
    :math:`P_{n}(\cdot)` denotes the Legendre polynomial of degree :math:`n`,
    and :math:`\Theta` the angle between :math:`(\theta, \phi)`
    and :math:`(\theta_\text{pw}, \phi_\text{pw})`.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of the spherical secondary source distribution.
    npw : (3,) array_like
        Unit vector (propagation direction) of plane wave.
    fs : int
        Sampling frequency in Hertz.
    max_order : int, optional
        Ambisonics order.
    c : float, optional
        Speed of sound in m/s.
    s2z : callable, optional
        Function transforming s-domain poles and zeros into z-domain,
        e.g. :func:`matchedz_zpk`, :func:`scipy.signal.bilinear_zpk`.

    Returns
    -------
    delay : float
        Overall delay in seconds.
    weight : float
        Overall weight.
    sos : list of numpy.ndarray
        Second-order section filters :func:`scipy.signal.sosfilt`.
    phaseshift : (N,) numpy.ndarray
        Phase shift in radians.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    """
    if max_order is None:
        max_order = util.max_order_spherical_harmonics(len(x0))
    if c is None:
        c = default.c

    x0 = util.asarray_of_rows(x0)
    npw = util.asarray_1d(npw)
    phi0, theta0, _ = util.cart2sph(*x0.T)
    phipw, thetapw, _ = util.cart2sph(*npw)
    phaseshift = np.arccos(np.dot(x0 / r0, -npw))

    delay = -r0 / c
    weight = 4 * np.pi / r0
    sos = []
    for m in range(max_order + 1):
        _, p, _ = besselap(m, norm='delay')
        s_zeros = np.zeros(m)
        s_poles = c / r0 * p
        s_gain = 1
        z_zeros, z_poles, z_gain = s2z(s_zeros, s_poles, s_gain, fs)
        sos.append(zpk2sos(z_zeros, z_poles, z_gain, pairing='nearest'))
    selection = util.source_selection_all(len(x0))
    return delay, weight, sos, phaseshift, selection, secondary_source_point(c)


def point_3d(x0, r0, xs, fs, max_order=None, c=None, s2z=matchedz_zpk):
    r"""Virtual point source by 3-dimensional NFC-HOA.

    .. math::

        D(\phi_0, s) =
        \frac{\e{\frac{s}{c}(r_0-r_\text{s})}}{4 \pi r_0 r_\text{s}}
        \sum_{n=0}^{N}
        (2n+1) P_{n}(\cos\Theta)
        \Big(\frac{s-\frac{c}{r_\text{s}}\sigma_0}{s-\frac{c}{r_0}\sigma_0}\Big)^\mu
        \prod_{l=1}^{\nu}
        \frac{(s-\frac{c}{r_\text{s}}\sigma_l)^2-(\frac{c}{r_\text{s}}\omega_l)^2}
        {(s-\frac{c}{r_0}\sigma_l)^2+(\frac{c}{r_0}\omega_l)^2}

    The driving function is represented in the Laplace domain,
    from which the recursive filters are designed.
    :math:`\sigma_l + \i\omega_l` denotes the complex roots of
    the reverse Bessel polynomial.
    The number of second-order sections is
    :math:`\nu = \big\lfloor\tfrac{|m|}{2}\big\rfloor`,
    whereas the number of first-order section :math:`\mu` is either 0 or 1
    for even and odd :math:`|m|`, respectively.
    :math:`P_{n}(\cdot)` denotes the Legendre polynomial of degree :math:`n`,
    and :math:`\Theta` the angle between :math:`(\theta, \phi)`
    and :math:`(\theta_\text{s}, \phi_\text{s})`.

    Parameters
    ----------
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    r0 : float
        Radius of the spherial secondary source distribution.
    xs : (3,) array_like
        Virtual source position.
    fs : int
        Sampling frequency in Hertz.
    max_order : int, optional
        Ambisonics order.
    c : float, optional
        Speed of sound in m/s.
    s2z : callable, optional
        Function transforming s-domain poles and zeros into z-domain,
        e.g. :func:`matchedz_zpk`, :func:`scipy.signal.bilinear_zpk`.

    Returns
    -------
    delay : float
        Overall delay in seconds.
    weight : float
        Overall weight.
    sos : list of numpy.ndarray
        Second-order section filters :func:`scipy.signal.sosfilt`.
    phaseshift : (N,) numpy.ndarray
        Phase shift in radians.
    selection : (N,) numpy.ndarray
        Boolean array containing only ``True`` indicating that
        all secondary source are "active" for NFC-HOA.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.td.synthesize()`.

    """
    if max_order is None:
        max_order = util.max_order_spherical_harmonics(len(x0))
    if c is None:
        c = default.c

    x0 = util.asarray_of_rows(x0)
    xs = util.asarray_1d(xs)
    phi0, theta0, _ = util.cart2sph(*x0.T)
    phis, thetas, rs = util.cart2sph(*xs)
    phaseshift = np.arccos(np.dot(x0 / r0, xs / rs))

    delay = (rs - r0) / c
    weight = 1 / r0 / rs
    sos = []
    for m in range(max_order + 1):
        _, p, _ = besselap(m, norm='delay')
        s_zeros = c / rs * p
        s_poles = c / r0 * p
        s_gain = 1
        z_zeros, z_poles, z_gain = s2z(s_zeros, s_poles, s_gain, fs)
        sos.append(zpk2sos(z_zeros, z_poles, z_gain, pairing='nearest'))
    selection = util.source_selection_all(len(x0))
    return delay, weight, sos, phaseshift, selection, secondary_source_point(c)


def driving_signals_25d(delay, weight, sos, phaseshift, signal):
    """Get 2.5-dimensional NFC-HOA driving signals.

    Parameters
    ----------
    delay : float
        Overall delay in seconds.
    weight : float
        Overall weight.
    sos : list of array_like
        Second-order section filters :func:`scipy.signal.sosfilt`.
    phaseshift : (N,) array_like
        Phase shift in radians.
    signal : (L,) array_like + float
        Excitation signal consisting of (mono) audio data and a sampling
        rate (in Hertz).  A `DelayedSignal` object can also be used.

    Returns
    -------
    `DelayedSignal`
        A tuple containing the delayed signals (in a `numpy.ndarray`
        with shape ``(L, N)``), followed by the sampling rate (in Hertz)
        and a (possibly negative) time offset (in seconds).

    """
    data, fs, t_offset = util.as_delayed_signal(signal)
    N = len(phaseshift)
    out = np.tile(np.expand_dims(sosfilt(sos[0], data), 1), (1, N))
    for m in range(1, len(sos)):
        modal_response = sosfilt(sos[m], data)[:, np.newaxis]
        out += modal_response * np.cos(m * phaseshift)
    return util.DelayedSignal(2 * weight * out, fs, t_offset + delay)


def driving_signals_3d(delay, weight, sos, phaseshift, signal):
    """Get 3-dimensional NFC-HOA driving signals.

    Parameters
    ----------
    delay : float
        Overall delay in seconds.
    weight : float
        Overall weight.
    sos : list of array_like
        Second-order section filters :func:`scipy.signal.sosfilt`.
    phaseshift : (N,) array_like
        Phase shift in radians.
    signal : (L,) array_like + float
        Excitation signal consisting of (mono) audio data and a sampling
        rate (in Hertz).  A `DelayedSignal` object can also be used.

    Returns
    -------
    `DelayedSignal`
        A tuple containing the delayed signals (in a `numpy.ndarray`
        with shape ``(L, N)``), followed by the sampling rate (in Hertz)
        and a (possibly negative) time offset (in seconds).

    """
    data, fs, t_offset = util.as_delayed_signal(signal)
    N = len(phaseshift)
    out = np.tile(np.expand_dims(sosfilt(sos[0], data), 1), (1, N))
    for m in range(1, len(sos)):
        modal_response = sosfilt(sos[m], data)[:, np.newaxis]
        out += (2 * m + 1) * modal_response * legendre(m, np.cos(phaseshift))
    return util.DelayedSignal(weight / 4 / np.pi * out, fs, t_offset + delay)



"""Compute SDM driving functions.

.. include:: math-definitions.rst

.. plot::
    :context: reset

    import matplotlib.pyplot as plt
    import numpy as np
    import sfs

    plt.rcParams['figure.figsize'] = 6, 6

    xs = -1.5, 1.5, 0
    # normal vector for plane wave:
    npw = sfs.util.direction_vector(np.radians(-45))
    f = 300  # Hz
    omega = 2 * np.pi * f

    grid = sfs.util.xyz_grid([-2, 2], [-2, 2], 0, spacing=0.02)

    array = sfs.array.linear(32, 0.2, orientation=[0, -1, 0])

    def plot(d, selection, secondary_source):
        p = sfs.fd.synthesize(d, selection, array, secondary_source, grid=grid)
        sfs.plot2d.amplitude(p, grid)
        sfs.plot2d.loudspeakers(array.x, array.n, selection * array.a, size=0.15)

"""

import numpy as np
from scipy.special import hankel2
from .. import util
from . import secondary_source_line, secondary_source_point


def line_2d(omega, x0, n0, xs, *, c=None):
    r"""Driving function for 2-dimensional SDM for a virtual line source.

    Parameters
    ----------
    omega : float
        Angular frequency of line source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs : (3,) array_like
        Position of line source.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    The secondary sources have to be located on the x-axis (y0=0).
    Derived from :cite:`Spors2009`, Eq.(9), Eq.(4).

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.sdm.line_2d(
            omega, array.x, array.n, xs)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = - 1j/2 * k * xs[1] / r * hankel2(1, k * r)
    selection = util.source_selection_all(len(x0))
    return d, selection, secondary_source_line(omega, c)


def plane_2d(omega, x0, n0, n=[0, 1, 0], *, c=None):
    r"""Driving function for 2-dimensional SDM for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of plane wave.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    n: (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    The secondary sources have to be located on the x-axis (y0=0).
    Derived from :cite:`Ahrens2012`, Eq.(3.73), Eq.(C.5), Eq.(C.11):

    .. math::

        D(\x_0,k) = k_\text{pw,y} \e{-\i k_\text{pw,x} x}

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.sdm.plane_2d(
            omega, array.x, array.n, npw)
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    n = util.normalize_vector(n)
    k = util.wavenumber(omega, c)
    d = k * n[1] * np.exp(-1j * k * n[0] * x0[:, 0])
    selection = util.source_selection_all(len(x0))
    return d, selection, secondary_source_line(omega, c)


def plane_25d(omega, x0, n0, n=[0, 1, 0], *, xref=[0, 0, 0], c=None):
    r"""Driving function for 2.5-dimensional SDM for a virtual plane wave.

    Parameters
    ----------
    omega : float
        Angular frequency of plane wave.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    n: (3,) array_like, optional
        Normal vector (traveling direction) of plane wave.
    xref : (3,) array_like, optional
        Reference point for synthesized sound field.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    The secondary sources have to be located on the x-axis (y0=0).
    Eq.(3.79) from :cite:`Ahrens2012`.

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.sdm.plane_25d(
            omega, array.x, array.n, npw, xref=[0, -1, 0])
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    n = util.normalize_vector(n)
    xref = util.asarray_1d(xref)
    k = util.wavenumber(omega, c)
    d = 4j * np.exp(-1j*k*n[1]*xref[1]) / hankel2(0, k*n[1]*xref[1]) * \
        np.exp(-1j*k*n[0]*x0[:, 0])
    selection = util.source_selection_all(len(x0))
    return d, selection, secondary_source_point(omega, c)


def point_25d(omega, x0, n0, xs, *, xref=[0, 0, 0], c=None):
    r"""Driving function for 2.5-dimensional SDM for a virtual point source.

    Parameters
    ----------
    omega : float
        Angular frequency of point source.
    x0 : (N, 3) array_like
        Sequence of secondary source positions.
    n0 : (N, 3) array_like
        Sequence of normal vectors of secondary sources.
    xs: (3,) array_like
        Position of virtual point source.
    xref : (3,) array_like, optional
        Reference point for synthesized sound field.
    c : float, optional
        Speed of sound.

    Returns
    -------
    d : (N,) numpy.ndarray
        Complex weights of secondary sources.
    selection : (N,) numpy.ndarray
        Boolean array containing ``True`` or ``False`` depending on
        whether the corresponding secondary source is "active" or not.
    secondary_source_function : callable
        A function that can be used to create the sound field of a
        single secondary source.  See `sfs.fd.synthesize()`.

    Notes
    -----
    The secondary sources have to be located on the x-axis (y0=0).
    Driving function from :cite:`Spors2010`, Eq.(24).

    Examples
    --------
    .. plot::
        :context: close-figs

        d, selection, secondary_source = sfs.fd.sdm.point_25d(
            omega, array.x, array.n, xs, xref=[0, -1, 0])
        plot(d, selection, secondary_source)

    """
    x0 = util.asarray_of_rows(x0)
    n0 = util.asarray_of_rows(n0)
    xs = util.asarray_1d(xs)
    xref = util.asarray_1d(xref)
    k = util.wavenumber(omega, c)
    ds = x0 - xs
    r = np.linalg.norm(ds, axis=1)
    d = 1/2 * 1j * k * np.sqrt(xref[1] / (xref[1] - xs[1])) * \
        xs[1] / r * hankel2(1, k * r)
    selection = util.source_selection_all(len(x0))
    return d, selection, secondary_source_point(omega, c)



""" Example for particle density visualization of sound sources """

import numpy as np
import matplotlib.pyplot as plt
import sfs

# simulation parameters
pw_angle = 45  # traveling direction of plane wave
xs = [0, 0, 0]  # source position
f = 300  # frequency

# angular frequency
omega = 2 * np.pi * f
# normal vector of plane wave
npw = sfs.util.direction_vector(np.radians(pw_angle))
# random grid for velocity
grid = [np.random.uniform(-3, 3, 40000), np.random.uniform(-3, 3, 40000), 0]


def plot_particle_displacement(title):
    # compute displacement
    X = grid + amplitude * sfs.fd.displacement(v, omega)
    # plot displacement
    plt.figure(figsize=(15, 15))
    plt.cla()
    sfs.plot2d.particles(X, facecolor='black', s=3, trim=[-3, 3, -3, 3])
    plt.axis('off')
    plt.title(title)
    plt.grid()
    plt.savefig(title + '.png')


# point source
v = sfs.fd.source.point_velocity(omega, xs, grid)
amplitude = 1.5e6
plot_particle_displacement('particle_displacement_point_source')

# line source
v = sfs.fd.source.line_velocity(omega, xs, grid)
amplitude = 1.3e6
plot_particle_displacement('particle_displacement_line_source')

# plane wave
v = sfs.fd.source.plane_velocity(omega, xs, npw, grid)
amplitude = 1e5
plot_particle_displacement('particle_displacement_plane_wave')



"""This example illustrates the synthesis of a sound figure.

The sound figure is defined by a grayscale PNG image. Various example
images are located in the "figures/" directory.

"""
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import sfs

dx = 0.10  # secondary source distance
N = 60  # number of secondary sources
pw_angle = [90, 45]  # traveling direction of plane wave
f = 2000  # frequency

# angular frequency
omega = 2 * np.pi * f

# normal vector of plane wave
npw = sfs.util.direction_vector(*np.radians(pw_angle))

# spatial grid
grid = sfs.util.xyz_grid([-3, 3], [-3, 3], 0, spacing=0.02)

# get secondary source positions
array = sfs.array.cube(N, dx)

# driving function for sound figure
figure = np.array(Image.open('figures/tree.png'))  # read image from file
figure = np.rot90(figure)  # turn 0deg to the top
d, selection, secondary_source = sfs.fd.wfs.soundfigure_3d(
    omega, array.x, array.n, figure, npw=npw)

# compute synthesized sound field
p = sfs.fd.synthesize(d, selection, array, secondary_source, grid=grid)

# plot and save synthesized sound field
plt.figure(figsize=(10, 10))
sfs.plot2d.amplitude(p, grid, xnorm=[0, -2.2, 0], cmap='BrBG', colorbar=False,
                     vmin=-1, vmax=1)
plt.title('Synthesized Sound Field')
plt.savefig('soundfigure.png')

# plot and save level of synthesized sound field
plt.figure(figsize=(12.5, 12.5))
im = sfs.plot2d.level(p, grid, xnorm=[0, -2.2, 0], vmin=-50, vmax=0,
                      colorbar_kwargs=dict(label='dB'))
plt.title('Level of Synthesized Sound Field')
plt.savefig('soundfigure_level.png')



"""Weights (tapering) for the driving function.

.. plot::
    :context: reset

    import sfs
    import matplotlib.pyplot as plt
    import numpy as np
    plt.rcParams['figure.figsize'] = 8, 3  # inch
    plt.rcParams['axes.grid'] = True

    active1 = np.zeros(101, dtype=bool)
    active1[5:-5] = True

    # The active part can wrap around from the end to the beginning:
    active2 = np.ones(101, dtype=bool)
    active2[30:-10] = False

"""
import numpy as np


def none(active):
    """No tapering window.

    Parameters
    ----------
    active : array_like, dtype=bool
        A boolean array containing ``True`` for active loudspeakers.

    Returns
    -------
    type(active)
        The input, unchanged.

    Examples
    --------
    .. plot::
        :context: close-figs

        plt.plot(sfs.tapering.none(active1))
        plt.axis([-3, 103, -0.1, 1.1])

    .. plot::
        :context: close-figs

        plt.plot(sfs.tapering.none(active2))
        plt.axis([-3, 103, -0.1, 1.1])

    """
    return active


def tukey(active, *, alpha):
    """Tukey tapering window.

    This uses a function similar to :func:`scipy.signal.tukey`, except
    that the first and last value are not zero.

    Parameters
    ----------
    active : array_like, dtype=bool
        A boolean array containing ``True`` for active loudspeakers.
    alpha : float
        Shape parameter of the Tukey window, see
        :func:`scipy.signal.tukey`.

    Returns
    -------
    (len(active),) `numpy.ndarray`
        Tapering weights.

    Examples
    --------
    .. plot::
        :context: close-figs

        plt.plot(sfs.tapering.tukey(active1, alpha=0), label='alpha = 0')
        plt.plot(sfs.tapering.tukey(active1, alpha=0.25), label='alpha = 0.25')
        plt.plot(sfs.tapering.tukey(active1, alpha=0.5), label='alpha = 0.5')
        plt.plot(sfs.tapering.tukey(active1, alpha=0.75), label='alpha = 0.75')
        plt.plot(sfs.tapering.tukey(active1, alpha=1), label='alpha = 1')
        plt.axis([-3, 103, -0.1, 1.1])
        plt.legend(loc='lower center')

    .. plot::
        :context: close-figs

        plt.plot(sfs.tapering.tukey(active2, alpha=0.3))
        plt.axis([-3, 103, -0.1, 1.1])

    """
    idx = _windowidx(active)
    alpha = np.clip(alpha, 0, 1)
    if alpha == 0:
        return none(active)
    # design Tukey window
    x = np.linspace(0, 1, len(idx) + 2)
    tukey = np.ones_like(x)
    first_part = x < alpha / 2
    tukey[first_part] = 0.5 * (
        1 + np.cos(2 * np.pi / alpha * (x[first_part] - alpha / 2)))
    third_part = x >= (1 - alpha / 2)
    tukey[third_part] = 0.5 * (
        1 + np.cos(2 * np.pi / alpha * (x[third_part] - 1 + alpha / 2)))
    # fit window into tapering function
    result = np.zeros(len(active))
    result[idx] = tukey[1:-1]
    return result


def kaiser(active, *, beta):
    """Kaiser tapering window.

    This uses :func:`numpy.kaiser`.

    Parameters
    ----------
    active : array_like, dtype=bool
        A boolean array containing ``True`` for active loudspeakers.
    alpha : float
        Shape parameter of the Kaiser window, see :func:`numpy.kaiser`.

    Returns
    -------
    (len(active),) `numpy.ndarray`
        Tapering weights.

    Examples
    --------
    .. plot::
        :context: close-figs

        plt.plot(sfs.tapering.kaiser(active1, beta=0), label='beta = 0')
        plt.plot(sfs.tapering.kaiser(active1, beta=2), label='beta = 2')
        plt.plot(sfs.tapering.kaiser(active1, beta=6), label='beta = 6')
        plt.plot(sfs.tapering.kaiser(active1, beta=8.6), label='beta = 8.6')
        plt.plot(sfs.tapering.kaiser(active1, beta=14), label='beta = 14')
        plt.axis([-3, 103, -0.1, 1.1])
        plt.legend(loc='lower center')

    .. plot::
        :context: close-figs

        plt.plot(sfs.tapering.kaiser(active2, beta=7))
        plt.axis([-3, 103, -0.1, 1.1])

    """
    idx = _windowidx(active)
    window = np.zeros(len(active))
    window[idx] = np.kaiser(len(idx), beta)
    return window


def _windowidx(active):
    """Return list of connected indices for window function.

    Note: Gaps within the active part are not allowed.

    """
    # find index where active loudspeakers begin (works for connected contours)
    if (active[0] and not active[-1]) or np.all(active):
        first_idx = 0
    else:
        first_idx = np.argmax(np.diff(active.astype(int))) + 1
    # shift generic index vector to get a connected list of indices
    idx = np.roll(np.arange(len(active)), -first_idx)
    # remove indices of inactive secondary sources
    return idx[:np.count_nonzero(active)]



# This is a configuration file that's used when opening the Jupyter notebooks
# in this directory.
# See https://nbviewer.jupyter.org/github/mgeier/python-audio/blob/master/plotting/matplotlib-inline-defaults.ipynb

c.InlineBackend.figure_formats = {'svg'}
c.InlineBackend.rc = {'figure.dpi': 96}



import numpy as np
from numpy.testing import assert_array_equal
import pytest
import sfs


def vectortypes(*coeffs):
    return [
        list(coeffs),
        tuple(coeffs),
        np.array(coeffs),
        np.array(coeffs).reshape(1, -1),
        np.array(coeffs).reshape(-1, 1),
    ]


def vector_id(vector):
    if isinstance(vector, np.ndarray):
        return 'array, shape=' + repr(vector.shape)
    return type(vector).__name__


@pytest.mark.parametrize('N, spacing, result', [
    (2, 1, sfs.array.SecondarySourceDistribution(
        x=[[0, -0.5, 0], [0, 0.5, 0]],
        n=[[1, 0, 0], [1, 0, 0]],
        a=[1, 1],
    )),
    (3, 1, sfs.array.SecondarySourceDistribution(
        x=[[0, -1, 0], [0, 0, 0], [0, 1, 0]],
        n=[[1, 0, 0], [1, 0, 0], [1, 0, 0]],
        a=[1, 1, 1],
    )),
    (3, 0.5, sfs.array.SecondarySourceDistribution(
        x=[[0, -0.5, 0], [0, 0, 0], [0, 0.5, 0]],
        n=[[1, 0, 0], [1, 0, 0], [1, 0, 0]],
        a=[0.5, 0.5, 0.5],
    )),
])
def test_linear_with_defaults(N, spacing, result):
    a = sfs.array.linear(N, spacing)
    assert a.x.dtype == np.float64
    assert a.n.dtype == np.float64
    assert a.a.dtype == np.float64
    assert_array_equal(a.x, result.x)
    assert_array_equal(a.n, result.n)
    assert_array_equal(a.a, result.a)


def test_linear_with_named_arguments():
    a = sfs.array.linear(N=2, spacing=0.5)
    assert_array_equal(a.x, [[0, -0.25, 0], [0, 0.25, 0]])
    assert_array_equal(a.n, [[1, 0, 0], [1, 0, 0]])
    assert_array_equal(a.a, [0.5, 0.5])


@pytest.mark.parametrize('center', vectortypes(-1, 0.5, 2), ids=vector_id)
def test_linear_with_center(center):
    a = sfs.array.linear(2, 1, center=center)
    assert_array_equal(a.x, [[-1, 0, 2], [-1, 1, 2]])
    assert_array_equal(a.n, [[1, 0, 0], [1, 0, 0]])
    assert_array_equal(a.a, [1, 1])


@pytest.mark.parametrize('orientation', vectortypes(0, -1, 0), ids=vector_id)
def test_linear_with_center_and_orientation(orientation):
    a = sfs.array.linear(2, 1, center=[0, 1, 2], orientation=orientation)
    assert_array_equal(a.x, [[-0.5, 1, 2], [0.5, 1, 2]])
